from .librrsimulator import libRRSimulator
from .csimulator import CSimulator


def sim_getter(model_file=None, model_str=None, sim_type="libRR"):
    """
    Convenience function to get a simulator object of a specific type.
    Allows you to pull a simulator object given a model file path.

    Note: This likely needs to be refactored but for now it works.

    Parameters
    ----------
    model_file : str, optional
        The path to the model file, at the moment only BNGL is expected
        but this can change in the future.
    model_str : str, optional
        Instead of the path to the model you can also supply the model
        string instead.
    sim_type : str, optional
        The name of the type of simulator object to get. At the moment only
        libRoadRunner type simulators are allowed. This will get updated
        as differenty types of simulators are added.

    Returns
    -------
    BNGSimulator
        A simulator object with an API that's supposed to be agnostic to the
        underlying simulator it's running.
    """
    if model_str is not None and model_file is None:
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile("w+") as model_file_obj:
            model_file_obj.write(model_str)
            model_file = model_file_obj.name
            if sim_type == "libRR":
                # need to go back to beginning of the file for this to work
                model_file_obj.seek(0)
                return libRRSimulator(model_file=model_file)
            elif sim_type == "cpy":
                return CSimulator(model_file=model_file, generate_network=True)
            else:
                print("simulator type {} not supported".format(sim_type))
    if model_file is not None:
        if sim_type == "libRR":
            return libRRSimulator(model_file=model_file)
        elif sim_type == "cpy":
            return CSimulator(model_file=model_file, generate_network=True)
        else:
            print("simulator type {} not supported".format(sim_type))



class BNGError(Exception):
    """Base class for all PyBNG exceptions."""

    pass


class BNGVersionError(BNGError):
    """Error related to mismatching required PyBNG."""

    def __init__(self, cur_version, req_version):
        self.message = f"Your PyBNG version: {cur_version} doesn't match required version: {req_version}. Try updating PyBNG with `pip install bionetgen --upgrade`"
        self.cur_version = cur_version
        self.req_version = req_version
        super().__init__(self.message)


class BNGPerlError(BNGError):
    """Error related to BNG2.pl/existence of perl."""

    def __init__(self):
        full_msg = "Perl doesn't seem to be installed, please install perl"
        full_msg += "We recommend strawberry perl (https://strawberryperl.com/) for Windows users."
        self.message = full_msg
        super().__init__(self.message)


class BNGParseError(BNGError):
    """Error related to parsing a BNGL file."""

    def __init__(
        self, bngl_path=None, message="There was an issue with parsing your BNGL file"
    ):
        self.path = bngl_path
        full_msg = f"There was an issue parsing BNGL file: {bngl_path}"
        if message is not None:
            full_msg += message
        self.message = full_msg
        super().__init__(self.message)


class BNGFileError(BNGError):
    """Error related to the BNGL file."""

    def __init__(self, bngl_path, message="There was an issue with your BNGL file"):
        self.path = bngl_path
        self.message = message
        super().__init__(self.message)


class BNGModelError(BNGError):
    """Error related to the BNG model object."""

    def __init__(self, model, message="There was an issue with your BNG model"):
        self.model = model
        full_msg = f"There was an issue with BNGL model: {self.model}\n"
        if message is not None:
            full_msg += message
        self.message = full_msg
        super().__init__(self.message)


class BNGRunError(BNGError):
    """Error related to running BNG2.pl."""

    def __init__(
        self,
        command,
        message="There was an issue with running BNG2.pl. There might be an issue with your model.",
        stdout=None,
        stderr=None,
    ):
        self.command = command
        self.stdout = stdout
        self.stderr = stderr
        full_msg = f"Tried to run command: {command}\n"
        full_msg += message + "\n"
        if stdout is not None:
            full_msg += f"Stdout was: {stdout}\n"
        if stderr is not None:
            full_msg += f"Stderr was: {stderr}\n"
        self.message = full_msg
        super().__init__(self.message)


class BNGCompileError(BNGError):
    """Error related to compiling C/Py file of BNG model."""

    def __init__(
        self,
        model,
        message="There was an issue compiling your BNG model, please make sure your bionetgen.conf file points to a compiled CVODE 2.6.0",
    ):
        self.model = model
        self.message = message
        super().__init__(self.message)



class BindingException(Exception):
    def __init__(self, value, combinations):
        self.value = value
        self.combinations = combinations

    def __str__(self):
        return repr(self.value)



from . import componentGroups
import argparse
import pprint
from collections import defaultdict
import itertools
from copy import copy
from bionetgen.atomizer.utils import readBNGXML

import functools
import marshal


def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = marshal.dumps([args, kwargs])
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]

    return memoizer


@memoize
def resolveEntry(dependencyGraph, moleculeSet):
    """
    resolve an entry to its basic components according to dependency graph
    >>> dependencyGraph = {'EGF_EGFR_2':[['EGF_EGFR','EGF_EGFR']],'EGF_EGFR':[['EGF','EGFR']],'EGFR':[],'EGF':[]}
    >>> resolveEntry(dependencyGraph, ['EGF_EGFR_2'])
    ['EGF', 'EGFR', 'EGF', 'EGFR']

    """
    if type(moleculeSet) == str:
        return [moleculeSet]
    if len(moleculeSet) == 1 and dependencyGraph[moleculeSet[0]] == []:
        return moleculeSet
    compositionList = []
    for molecule in moleculeSet:
        if len(dependencyGraph[molecule]) == 0:
            compositionList.append(molecule)
        else:
            compositionList.extend(
                resolveEntry(dependencyGraph, dependencyGraph[molecule][0])
            )
    return compositionList


class ModelLearning:
    def __init__(self, fileName, rawFileName=None):
        self.molecules, self.rules, _ = readBNGXML.parseXML(fileName)
        (
            self.dependencies,
            self.patternXreactions,
            _,
            _,
        ) = componentGroups.getContextRequirements(fileName, collapse=False)
        self.transposePatternsReactions()
        self.reverseDependencies = componentGroups.reverseContextDict(self.dependencies)
        self.moleculeMotifDict, self.motifMoleculeDict = self.classifyPairsByMotif(
            self.reverseDependencies
        )

        if rawFileName:
            self.processRawSBML(rawFileName)
        else:
            self.rawRules = {}

    def processRawSBML(self, inputfile):
        _, rawrules, _ = readBNGXML.parseXML(inputfile)
        self.rawRules = {x[0].label: x[0] for x in rawrules}

    def transposePatternsReactions(self):
        self.reactionsXpatterns = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        for product in self.patternXreactions:
            for reactionCenter in self.patternXreactions[product]:
                for contextcomponent in self.patternXreactions[product][reactionCenter]:
                    for contextState in self.patternXreactions[product][reactionCenter][
                        contextcomponent
                    ]:
                        for reaction in self.patternXreactions[product][reactionCenter][
                            contextcomponent
                        ][contextState]:
                            self.reactionsXpatterns[reaction][product][
                                reactionCenter
                            ].append((contextcomponent, contextState))

    def classifyPairsByMotif(self, reverseDependencies):
        """
        recives a dict arranged by molecule->componentPair->dependency

        organizes molecule pairs by a concept called 'motif'
        """
        motifDependencies = defaultdict(lambda: defaultdict(list))
        motifDependencies2 = defaultdict(lambda: defaultdict(list))

        for molecule in reverseDependencies:
            for moleculePair in reverseDependencies[molecule]:
                motifDependencies[molecule][frozenset(moleculePair)].append(
                    reverseDependencies[molecule][moleculePair]
                )

        for molecule in motifDependencies:
            for moleculePair in motifDependencies[molecule]:
                requirementPair = tuple(
                    sorted(motifDependencies[molecule][moleculePair])
                )
                motifDependencies2[requirementPair][molecule].append(moleculePair)

        return motifDependencies, motifDependencies2

    def getMotifFromPair(self, molecule, component1, component2):
        return self.moleculeMotifDict[molecule][frozenset([component1, component2])]

    def getParticipatingReactions(self, molecule, componentPair, reactionDictionary):
        correlationList = {}
        for moleculeName in reactionDictionary:
            if moleculeName.startswith(molecule + "%"):
                for component in reactionDictionary[moleculeName]:
                    if component[0] in componentPair and (
                        component[1] == 1 or component[2] not in ["0", 0, ""]
                    ):
                        for componentComplement in [
                            x
                            for x in reactionDictionary[moleculeName][component]
                            if x in componentPair
                        ]:
                            correlationList[
                                (component[0], componentComplement)
                            ] = reactionDictionary[moleculeName][component][
                                componentComplement
                            ]
        return correlationList

    def getPairsFromMotif(self, motif1, motif2, excludedComponents):
        moleculeCorrelationList = defaultdict(dict)
        for element in self.motifMoleculeDict:
            if motif1 in element and motif2 in element:
                if motif1 == motif2 and len([x for x in element if x == motif1]) < 2:
                    continue
                for molecule in self.motifMoleculeDict[element]:
                    correlationList = {}
                    if len(self.motifMoleculeDict[element][molecule]) > 0:
                        for componentPair in self.motifMoleculeDict[element][molecule]:
                            if not any(x in excludedComponents for x in componentPair):
                                correlationList[
                                    componentPair
                                ] = self.getParticipatingReactions(
                                    molecule, componentPair, self.patternXreactions
                                )
                    moleculeCorrelationList[molecule].update(correlationList)
        return dict(moleculeCorrelationList)

    def analyzeRedundantBonds(self, assumptions):
        """
        Analyzes a system of molecules with redundant bonds between them (more than one path between any two nodes in the system). The function
        attemps to score the bonds by looking out for partial competition relationships (e.g. the presence of one component
        excludes the activation of another, but in the other direction we see independence) which are less likely to occur than a fully independence
        relationship. The method will thus nominate such edges for deletion if the overall systems  still forms
        a fully connected graph after the bond removal.
        """

        def fullyConnectedGraph(nodes, edges):
            """
            Lazy implementation. This only works if there is one common elements to all subgraphs
            """
            if edges == []:
                return False
            tmpNodeList = [set(x) for x in edges]
            superGraph = set.intersection(*map(set, edges))
            if len(superGraph) > 0:
                return True
            return False

        conserveBonds = []

        deleteBonds = {}
        for redundantBondSet in assumptions:
            allBonds = [sorted(x) for x in itertools.combinations(redundantBondSet, 2)]
            conserveBonds = []
            for molecule in redundantBondSet:
                for x in itertools.combinations(
                    [x for x in redundantBondSet if x != molecule], 2
                ):
                    contextMotif = self.getMotifFromPair(
                        molecule, x[0].lower(), x[1].lower()
                    )
                    if (
                        "independent" in contextMotif
                        and not (
                            "requirement" in contextMotif
                            or "nullrequirement" in contextMotif
                        )
                    ) or set(contextMotif) == set(["requirement", "nullrequirement"]):
                        conserveBonds.append(sorted([molecule, x[0]]))
                        conserveBonds.append(sorted([molecule, x[1]]))
            if fullyConnectedGraph(redundantBondSet, conserveBonds):
                deleteBonds[redundantBondSet] = [
                    x for x in allBonds if x not in conserveBonds
                ]
        return deleteBonds

    def scoreHypotheticalBonds(self, assumptions):
        """
        TODO: we need some way to evaluate the confidence in a bond based on context information
        """
        pass

    def getDifference(self, pattern1, pattern2, translator):
        if pattern1 not in translator or pattern2 not in translator:
            return None

        species1 = translator[pattern1]
        species2 = translator[pattern2]
        species1.sort()
        species2.sort()
        componentDifference = []
        for molecule1, molecule2 in zip(species1.molecules, species2.molecules):
            for component1, component2 in zip(
                molecule1.components, molecule2.components
            ):
                if (
                    len(component1.bonds) != len(component2.bonds)
                    or component1.activeState != component2.activeState
                ):
                    componentDifference.append(component1.name)
        return componentDifference

    def processContextMotifInformation(self, assumptionList, database):
        def getClassification(keys, translator):
            flags = [key in [x.lower() for x in translator.keys()] for key in keys]
            if all(flags):
                return "binding-binding"
            elif any(flags):
                return "modification-binding"
            return "modification-modification"

        motifInformationDict = self.getContextMotifInformation()

        motifFinalLog = defaultdict(set)
        motifReactionDefinitions = {}
        for motifInformation in motifInformationDict:
            # iterate over motifs that are known to be problematic
            if motifInformation in [
                frozenset(["nullrequirement", "independent"]),
                frozenset(["requirement", "independent"]),
            ]:

                for molecule in motifInformationDict[motifInformation]:
                    if len(motifInformationDict[motifInformation][molecule]) == 0:
                        continue
                    # if the candidate definitions for a given compound are related to a molecule with problematic motifs
                    for assumption in (
                        x
                        for x in assumptionList
                        for y in eval(x[3][1])
                        for z in y
                        if molecule in z
                    ):
                        candidates = eval(assumption[1][1])
                        alternativeCandidates = eval(assumption[2][1])
                        original = eval(assumption[3][1])
                        # further confirm that the change is about the pair of interest
                        # by iterating over all candidates and comparing one by one
                        for candidate in candidates:
                            for alternativeCandidate in alternativeCandidates:
                                difference = [
                                    x
                                    for x in candidate
                                    if x not in alternativeCandidate
                                ]
                                difference.extend(
                                    [
                                        x
                                        for x in alternativeCandidate
                                        if x not in candidate
                                    ]
                                )
                                # if we are referencing a molecule that is not about this particular context change
                                # dont store it in the motif/species table, just keep information about the motif alone
                                localAnalysisFlag = True
                                if not any(
                                    [
                                        molecule
                                        in database.prunnedDependencyGraph[x][0]
                                        if len(database.prunnedDependencyGraph[x]) > 0
                                        else molecule in x
                                        for x in difference
                                    ]
                                ):
                                    localAnalysisFlag = False
                                #    continue
                                if localAnalysisFlag:
                                    # get those elements that differ between the two candidates and that correspond to the current <molecule> being analyzed
                                    difference = [
                                        x
                                        for x in candidate
                                        if x not in alternativeCandidate
                                        and resolveEntry(
                                            database.prunnedDependencyGraph, [x]
                                        )[0]
                                        == molecule
                                    ]
                                    alternativeDifference = [
                                        x
                                        for x in alternativeCandidate
                                        if x not in candidate
                                        and molecule
                                        in resolveEntry(
                                            database.prunnedDependencyGraph, [x]
                                        )[0]
                                    ]

                                    # get the difference patterns for the two species
                                    if not difference or not alternativeDifference:
                                        continue
                                    componentDifference = self.getDifference(
                                        difference[0],
                                        alternativeDifference[0],
                                        database.translator,
                                    )

                                # make sure that the current motif candidate intersects with the difference pattern
                                for keys in motifInformationDict[motifInformation][
                                    molecule
                                ]:
                                    if localAnalysisFlag and any(
                                        key in componentDifference for key in keys
                                    ):
                                        motifFinalLog[
                                            "{0}({1})".format(molecule, ", ".join(keys))
                                        ].add(assumption[0])
                                    classification = getClassification(
                                        keys, database.translator
                                    )
                                    if classification not in motifReactionDefinitions:
                                        motifReactionDefinitions[classification] = {}
                                    motifReactionDefinitions[classification][
                                        "{0}({1})".format(molecule, ", ".join(keys))
                                    ] = (
                                        motifInformation,
                                        motifInformationDict[motifInformation][
                                            molecule
                                        ][keys],
                                    )

        # pprint.pprint(dict(motifFinalLog))
        # pprint.pprint(motifReactionDefinitions)
        return motifFinalLog, motifReactionDefinitions

    def getContextMotifInformation(self):
        """
        returns the reactions in the system classified by context-component motif pairs. e.g. a requirement,nullrequirement
        motif pair is a pair of components A->B such that B needs for A to be activated to activate, whereas A needs for B
        to be inactivated to activate.
        """
        relationshipCombinations = itertools.combinations(
            ["independent", "requirement", "nullrequirement", "exclusion"], 2
        )
        motifDictionary = {}

        for relCombi in relationshipCombinations:
            motifDictionary[frozenset(relCombi)] = self.getPairsFromMotif(
                relCombi[0], relCombi[1], []
            )

        for requirementClass in [
            "independent",
            "requirement",
            "nullrequirement",
            "exclusion",
        ]:
            motifDictionary[
                frozenset([requirementClass, requirementClass])
            ] = self.getPairsFromMotif(requirementClass, requirementClass, [])
        return motifDictionary

    def getComplexReactions(self, threshold=2):
        complexRules = []
        for rule in self.rules:
            if (
                len(
                    [
                        x
                        for x in rule[0].actions
                        if x.action not in ["ChangeCompartment"]
                    ]
                )
                >= threshold
            ):
                complexRules.append(rule)
        return complexRules

    def analyzeComplexReactions(self, threshold=2):
        def getActionableComponentPartners(actions, molecule):

            actionableComponents = []
            for action in actions:
                if action[1] and action[1] in molecule.lower():
                    actionableComponents.append(action[2])
                if action[2] and action[2] in molecule.lower():
                    actionableComponents.append(action[1])
            return actionableComponents

        for reaction in self.getComplexReactions():
            # analyze reactions with cis-allostery (e.g. two actions sites are on the same molecule)
            if (
                len(
                    [
                        x
                        for x in self.reactionsXpatterns[reaction[0].label]
                        if len(self.reactionsXpatterns[reaction[0].label][x]) > 1
                    ]
                )
                == 0
            ):
                continue
            print("----------")
            if reaction[0].label in self.rawRules:
                print(str(self.rawRules[reaction[0].label]))
                print(">>>>>>")
                print(str(reaction[0]))
            else:
                print(str(reaction[0].label))
            # print(str(reaction[0]))
            resolvedActions = []
            print("Actions:")
            changeFlag = 0
            for action in reaction[0].actions:
                molecule1 = (
                    reaction[-3]["_".join(action.site1.split("_")[:-1])]
                    if action.site1
                    else ""
                )
                molecule2 = (
                    reaction[-3]["_".join(action.site2.split("_")[:-1])]
                    if action.site2
                    else ""
                )
                site1 = reaction[-3][action.site1] if action.site1 else ""
                site2 = reaction[-3][action.site2] if action.site2 else ""
                print(
                    "\t{0}= {1}({2}), {3}({4})".format(
                        action.action, molecule1, site1, molecule2, site2
                    )
                )
                if action.action == "DeleteBond":
                    changeFlag = 1
                resolvedActions.append([action.action, site1, site2])

            print("Context:")
            for reactionCenter in self.reactionsXpatterns[reaction[0].label]:
                # cys rules

                if len(self.reactionsXpatterns[reaction[0].label][reactionCenter]) > 1:
                    for state in self.reactionsXpatterns[reaction[0].label][
                        reactionCenter
                    ]:
                        # we will focus on statechange actions for now
                        if state[2] not in [""]:
                            # print(self.patternXreactions[reactionCenter])
                            actionableComponents = getActionableComponentPartners(
                                [
                                    x
                                    for x in resolvedActions
                                    if x[0] in ["AddBond", "DeleteBond"]
                                ],
                                reactionCenter,
                            )
                            for component in actionableComponents:
                                print(
                                    "\treaction center <{0}>, context <{1}> in molecule <{2}>:".format(
                                        component, state[0], reactionCenter
                                    )
                                )
                                print(
                                    "\t",
                                    {
                                        x: dict(
                                            self.patternXreactions[reactionCenter][
                                                (component, changeFlag, "")
                                            ][x]
                                        )
                                        for x in self.patternXreactions[reactionCenter][
                                            (component, changeFlag, "")
                                        ]
                                        if x in [state[0], state[0].lower()]
                                    },
                                )

            print("+++++++++")


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument("-i", "--input", type=str, help="sbml file")
    parser.add_argument("-r", "--raw", type=str, help="raw sbml file")
    parser.add_argument("-t", "--tests", action="store_true", help="run unit tests")
    return parser


def runTests():
    import doctest

    doctest.testmod()


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()

    if namespace.tests:
        runTests()
        exit()
    inputFile = namespace.input

    modelLearning = ModelLearning(namespace.input, namespace.raw)
    # print(modelLearning.getMotifFromPair('EGFR','grb2','shc'))
    # print(modelLearning.getMotifFromPair('Shc','grb2','egfr'))
    # modelLearning.analyzeComplexReactions()

    # for rule in complexRules:
    #    print(str(rule[0]))

    relationshipCombinations = itertools.combinations(
        ["independent", "requirement", "nullrequirement", "exclusion"], 2
    )
    motifDictionary = {}
    for relCombi in relationshipCombinations:
        motifDictionary[relCombi] = modelLearning.getPairsFromMotif(
            relCombi[0], relCombi[1], ["imod"]
        )
        if len(motifDictionary[relCombi]) > 0:
            print(
                relCombi,
                {
                    x: len(motifDictionary[relCombi][x])
                    for x in motifDictionary[relCombi]
                },
            )
    for requirementClass in [
        "independent",
        "requirement",
        "nullrequirement",
        "exclusion",
    ]:
        motifDictionary[
            (requirementClass, requirementClass)
        ] = modelLearning.getPairsFromMotif(
            requirementClass, requirementClass, ["imod"]
        )
        if len(motifDictionary[(requirementClass, requirementClass)]) > 0:
            print(
                (requirementClass, requirementClass),
                {
                    x: len(motifDictionary[(requirementClass, requirementClass)][x])
                    for x in motifDictionary[(requirementClass, requirementClass)]
                },
            )
    print(modelLearning.getPairsFromMotif("independent", "requirement", ["imod"]))
    print("---")
    # print(modelLearning.getPairsFromMotif('independent','nullrequirement',['imod']))



import os
import numpy as np

from bionetgen.core.utils.logging import BNGLogger


class BNGResult:
    """
    Class that loads in gdat/cdat/scan files

    Usage: BNGResult(path="/path/to/folder") OR
           BNGResult(direct_path="/path/to/file.gdat")

    Arguments
    ---------
    path : str
        path that points to a folder containing files to be
        loaded by the class
    direct_path : str
        path that directly points to a file to load

    Methods
    -------
    load(fpath)
        loads in the direct path to the file and returns
        numpy.recarray
    """

    def __init__(self, path=None, direct_path=None, app=None):
        self.app = app
        self.logger = BNGLogger(app=self.app)
        self.logger.debug(
            "Setting up BNGResult object", loc=f"{__file__} : BNGResult.__init__()"
        )
        # defaults
        self.process_return = None
        self.output = None
        # TODO Make it so that with path you can supply an
        # extension or a list of extensions to load in
        self.gdats = {}
        self.cdats = {}
        self.scans = {}
        self.cnames = {}
        self.snames = {}
        self.gnames = {}
        if direct_path is not None:
            path, fname = os.path.split(direct_path)
            fnoext, fext = os.path.splitext(fname)
            self.direct_path = direct_path
            self.file_name = fnoext
            self.file_extension = fext
            self.gnames[fnoext] = direct_path
            self.gdats[fnoext] = self.load(direct_path)
        elif path is not None:
            # TODO change this pattern so that each method
            # is stand alone and usable.
            self.path = path
            self.find_dat_files()
            self.load_results()
        else:
            self.logger.info(
                "BNGResult needs either a path or a direct path kwarg to load gdat/cdat/scan files from",
                loc=f"{__file__} : BNGResult.__init__()",
            )

    def __repr__(self) -> str:
        s = f"gdats from {len(self.gdats)} models: "
        for r in self.gdats.keys():
            s += f"{r} "
        if len(self.cdats) > 0:
            s += f"\ncdats from {len(self.cdats)} models: "
            for r in self.cdats.keys():
                s += f"{r} "
        if len(self.scans) > 0:
            s += f"\nscans from {len(self.scans)} models: "
            for r in self.scans.keys():
                s += f"{r} "
        return s

    def __getitem__(self, key):
        if isinstance(key, int):
            k = list(self.gdats.keys())[key]
            it = self.gdats[k]
        else:
            it = self.gdats[key]
        return it

    def __iter__(self):
        return self.gdats.__iter__()

    def load(self, fpath):
        self.logger.debug(f"Loading file {fpath}", loc=f"{__file__} : BNGResult.load()")
        path, fname = os.path.split(fpath)
        fnoext, fext = os.path.splitext(fname)
        if fext == ".gdat" or fext == ".cdat":
            return self._load_dat(fpath)
        elif fext == ".scan":
            return self._load_scan(fpath)
        else:
            self.logger.info(
                "BNGResult doesn't know the file type of {}".format(fpath),
                loc=f"{__file__} : BNGResult.load()",
            )
            return None

    def _load_scan(self, fpath):
        return self._load_dat(fpath)

    def find_dat_files(self):
        self.logger.debug(
            f"Scanning for valid files in folder {self.path}",
            loc=f"{__file__} : BNGResult.find_dat_files()",
        )
        files = os.listdir(self.path)
        ext = "gdat"
        gdat_files = filter(lambda x: x.endswith(f".{ext}"), files)
        for dat_file in gdat_files:
            name = dat_file.replace(f".{ext}", "")
            self.gnames[name] = dat_file

        ext = "cdat"
        cdat_files = filter(lambda x: x.endswith(f".{ext}"), files)
        for dat_file in cdat_files:
            name = dat_file.replace(f".{ext}", "")
            self.cnames[name] = dat_file

        ext = "scan"
        scan_files = filter(lambda x: x.endswith(f".{ext}"), files)
        for dat_file in scan_files:
            name = dat_file.replace(f".{ext}", "")
            self.snames[name] = dat_file

    def load_results(self):
        self.logger.debug(
            f"Loading results from {self.path}",
            loc=f"{__file__} : BNGResult.load_results()",
        )
        # load gdat files
        for name in self.gnames:
            gdat_path = os.path.join(self.path, self.gnames[name])
            self.gdats[name] = self.load(gdat_path)
        # load gdat files
        for name in self.cnames:
            cdat_path = os.path.join(self.path, self.cnames[name])
            self.cdats[name] = self.load(cdat_path)
        # load scan files
        for name in self.snames:
            scan_path = os.path.join(self.path, self.snames[name])
            self.scans[name] = self.load(scan_path)

    def _load_dat(self, path, dformat="f8"):
        """
        This function takes a path to a gdat/cdat file as a string and loads that
        file into a numpy structured array, including the correct header info.
        TODO: Add link

        Optional argument allows you to set the data type for every column. See
        numpy dtype/data type strings for what's allowed. TODO: Add link
        """
        # First step is to read the header,
        # we gotta open the file and pull that line in
        with open(path, "r") as f:
            header = f.readline()
        # Ensure the header info is actually there
        # TODO: Transition to BNGErrors and logging
        assert header.startswith("#"), "No header line that starts with #"
        # Now turn it into a list of names for our struct array
        header = header.replace("#", "")
        headers = header.split()
        # For a magical reason this is how numpy.loadtxt wants it,
        # in tuples passed as a dictionary with names/formats as keys
        names = tuple(headers)
        formats = tuple([dformat for i in range(len(headers))])
        # return the loadtxt result as a record array
        # which is similar to pandas data format without the helper functions
        return np.rec.array(
            np.loadtxt(path, dtype={"names": names, "formats": formats})
        )



import readBNGXML
import argparse
import yaml


def normalizeMatchingMoleculeNamespace(molecule, moleculeSettings):
    # we have to change the name
    if "newname" in moleculeSettings:
        molecule.name = moleculeSettings["newname"]
    if "components" in moleculeSettings:
        for componentSettings in moleculeSettings["components"]:
            matchingComponents = [
                x for x in molecule.components if x.name == componentSettings["name"]
            ]
            for matchingComponent in matchingComponents:
                if "newname" in componentSettings:
                    matchingComponent.name = componentSettings["newname"]
                for stateSettings in componentSettings["states"]:
                    if stateSettings["name"] in matchingComponent.states:
                        idx = matchingComponent.states.index(stateSettings["name"])
                        if "newname" in stateSettings:
                            matchingComponent.states[idx] = stateSettings["newname"]
                            if matchingComponent.activeState == stateSettings["name"]:
                                matchingComponent.activeState = stateSettings["newname"]


def normalizeMoleculeNamespace(molecule, moleculeSettings):
    # molecule name matches
    if molecule.name == moleculeSettings["name"]:
        normalizeMatchingMoleculeNamespace(molecule, moleculeSettings)
    for component in molecule.components:
        if component.name == moleculeSettings["name"].lower():
            if "newname" in moleculeSettings:
                component.name = moleculeSettings["newname"].lower()


def normalizeNamespace(bnglNamespace, normalizationSettings):
    """Changes the namespace of a bngl file

    Keyword arguments:
    bnglNamespace -- The current molecule, rule etc structures
    normalizationSettings -- A dictionary containing those structures whose name we will change
    """

    # change molecule namespace
    for moleculeSettings in normalizationSettings["molecules"]:
        for molecule in bnglNamespace["molecules"]:
            normalizeMoleculeNamespace(molecule, moleculeSettings)

    # change seed species namespace
    for seedspecies in bnglNamespace["seedspecies"]:
        for molecule in seedspecies["structure"].molecules:
            for moleculeSettings in normalizationSettings["molecules"]:
                normalizeMoleculeNamespace(molecule, moleculeSettings)

    # change observable namespace
    for observables in bnglNamespace["observables"]:
        for pattern in observables[2]:
            for molecule in pattern.molecules:
                for moleculeSettings in normalizationSettings["molecules"]:
                    normalizeMoleculeNamespace(molecule, moleculeSettings)

    # change rule namespace
    for ruleDescription in bnglNamespace["rules"]:
        for reactant in ruleDescription[0].reactants:
            for molecule in reactant.molecules:
                for moleculeSettings in normalizationSettings["molecules"]:
                    normalizeMoleculeNamespace(molecule, moleculeSettings)
        for product in ruleDescription[0].products:
            for molecule in product.molecules:
                for moleculeSettings in normalizationSettings["molecules"]:
                    normalizeMoleculeNamespace(molecule, moleculeSettings)


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-n", "--normalize", type=str, help="normalization settings file", required=True
    )
    return parser


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()
    with open(namespace.normalize) as f:
        normalizationSettings = yaml.load(f)

    for model in normalizationSettings["model"]:
        bnglNamespace = readBNGXML.parseFullXML(model["name"])

        normalizeNamespace(bnglNamespace, model)

        bnglString = readBNGXML.createBNGLFromDescription(bnglNamespace)
        with open(model["name"] + "_norm.bngl", "w") as f:
            f.write(bnglString)



from utils import readBNGXML
import argparse
import collections
import stateTransitionDiagram as std
import xlwt
import yaml
import arial10
import os
import fnmatch
from collections import defaultdict


def getValidFiles(directory, extension):
    """
    Gets a list of bngl files that could be correctly translated in a given 'directory'
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, "*.{0}".format(extension)):
            matches.append(os.path.join(root, filename))
    return matches


class FitSheetWrapper(object):
    """Try to fit columns to max size of any entry.
    To use, wrap this around a worksheet returned from the
    workbook's add_sheet method, like follows:

        sheet = FitSheetWrapper(book.add_sheet(sheet_name))

    The worksheet interface remains the same: this is a drop-in wrapper
    for auto-sizing columns.
    """

    def __init__(self, sheet):
        self.sheet = sheet
        self.widths = dict()

    def write(self, r, c, label="", *args, **kwargs):
        self.sheet.write(r, c, label, *args, **kwargs)
        width = arial10.fitwidth(label)
        if width > self.widths.get(c, 0):
            self.widths[c] = int(width)
            self.sheet.col(c).width = int(width)

    def __getattr__(self, attr):
        return getattr(self.sheet, attr)


class PrettyDefaultDict(collections.defaultdict):
    __repr__ = dict.__repr__


def getParameterDictionary(bnglNamespace):

    parameterDict = PrettyDefaultDict(lambda: PrettyDefaultDict(set))
    simpleParameterDict = collections.defaultdict(set)

    (
        labels,
        centers,
        contexts,
        products,
        atomicArrays,
        actions,
        doubleAction,
    ) = std.extractCenterContext(bnglNamespace["rules"])

    for label, center, product, rule in zip(
        labels, centers, products, bnglNamespace["rules"]
    ):
        for pattern in product:
            for element in pattern:
                for molecule in element.split("."):
                    for rate in rule[0].rates:
                        value = (
                            bnglNamespace["parameters"][rate]
                            if rate in bnglNamespace["parameters"]
                            else "nma"
                        )
                        moleculeName = molecule.split("(")[0].split("%")[0]
                        component = molecule.split("(")[1].strip(")")
                        parameterDict[moleculeName.lower()][component].add(value)
                        simpleParameterDict[molecule].add(value)

    return parameterDict


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument("-i", "--input", type=str, help="settings file", required=True)
    return parser


def remove_empty_keys(d):
    for k in d.keys():
        if not d[k]:
            del d[k]


def getDifferences(modelNameList, parameterSpace):
    keys = collections.Counter([y.lower() for x in parameterSpace for y in x])

    rowIdx = 0
    parameterDict = {x: defaultdict(set) for x in keys}
    parameterModelDict = {x: defaultdict(list) for x in keys}
    for molecule in keys:
        if keys[molecule] <= 1:
            continue
        rowIdx = -1
        for midx, model in enumerate(parameterSpace):
            # write modelName
            if len(model[molecule]) == 0:
                continue
            rowIdx += 1

            for component in model[molecule]:
                if "!" in component:
                    componentTxt = component.split("!")[0] + ("(association)")
                elif "~0" in component:
                    componentTxt = component.split("~")[0] + ("(repression)")
                elif "~" in component:
                    componentTxt = component.split("~")[0] + ("(activation)")
                else:
                    componentTxt = component + ("(dissociation)")

                if (
                    all(
                        [
                            len(x.intersection(model[molecule][component])) == 0
                            for x in parameterDict[molecule][componentTxt]
                        ]
                    )
                    or len(parameterDict[molecule][componentTxt]) == 0
                ):
                    parameterDict[molecule][componentTxt].add(
                        frozenset(model[molecule][component])
                    )
                    parameterModelDict[molecule][componentTxt].append(
                        modelNameList[midx]
                    )
            removeKeys = []
        for component in parameterDict[molecule]:
            if len(parameterDict[molecule][component]) <= 1:
                removeKeys.append(component)
        for component in removeKeys:
            del parameterDict[molecule][component]
            if component in parameterModelDict[molecule]:
                parameterModelDict[molecule][component]

    remove_empty_keys(parameterDict)
    remove_empty_keys(parameterModelDict)

    import pprint

    print("---")
    pprint.pprint(parameterDict)


def ExcelOutput(modelNameList, parameterSpace):
    keys = collections.Counter([y.lower() for x in parameterSpace for y in x])
    wb = xlwt.Workbook()

    ws = FitSheetWrapper(wb.add_sheet("Units"))
    unitColumn = []
    rowIdx = 0

    for midx, element in enumerate(modelNameList):
        modelName = element.split(".")[:-1]
        modelName = ".".join(modelName)
        ymlName = modelName + ".bngl.yml"

        try:
            with open(ymlName, "r") as f:
                annotationDict = yaml.load(f)
        except IOError:
            continue
        ws.write(midx + 1, 0, modelName)
        for unit in annotationDict["units"]:
            if unit not in unitColumn:
                unitColumn.append(unit)
                ws.write(0, unitColumn.index(unit) + 1, unit)
            # for instance in annotationDict['units'][unit]:
            instance = annotationDict["units"][unit][0]
            if instance["exponent"] == 1:
                ws.write(
                    rowIdx + 1,
                    unitColumn.index(unit) + 1,
                    "{0} ({1})".format(instance["multiplier"], instance["name"]).decode(
                        "utf-8"
                    ),
                )
            else:
                ws.write(
                    rowIdx + 1,
                    unitColumn.index(unit) + 1,
                    "{0}^{2} ({1})".format(
                        instance["multiplier"], instance["name"], instance["exponent"]
                    ).decode("utf-8"),
                )
        rowIdx += 1
    for molecule in keys:
        if keys[molecule] <= 1:
            continue

        sheetName = molecule if len(molecule) <= 31 else molecule[0:31]
        ws = FitSheetWrapper(wb.add_sheet(sheetName))
        componentColumn = []
        rowIdx = -1
        for midx, model in enumerate(parameterSpace):
            # write modelName
            if len(model[molecule]) == 0:
                continue
            rowIdx += 1
            ws.write(rowIdx + 1, 0, modelNameList[midx])
            for component in model[molecule]:
                if component not in componentColumn:
                    componentColumn.append(component)
                    if "!" in component:
                        componentTxt = component.split("!")[0] + ("(association)")
                    elif "~0" in component:
                        componentTxt = component.split("~")[0] + ("(repression)")
                    elif "~" in component:
                        componentTxt = component.split("~")[0] + ("(activation)")
                    else:
                        componentTxt = component + ("(dissociation)")

                    ws.write(0, componentColumn.index(component) + 1, componentTxt)
                data = "/".join(model[molecule][component])
                ws.write(rowIdx + 1, componentColumn.index(component) + 1, data)

    wb.save("19_151_48.xls")


if __name__ == "__main__":
    # parser = defineConsole()
    # namespace = parser.parse_args()
    # inputFile = namespace.input
    # modelNameList = getValidFiles('egfr','xml')
    # modelNameList = getValidFiles('curated','xml')

    modelNameList = getValidFiles("tmp", "xml")
    # modelName = ['egfr/output151.xml']
    parameterSpace = []
    for element in modelNameList:
        parameterSpace.append(getParameterDictionary(readBNGXML.parseFullXML(element)))

    ExcelOutput(modelNameList, parameterSpace)
    getDifferences(modelNameList, parameterSpace)
    """
    with open('19_151_48.csv','wb') as f:
        writer = csv.DictWriter(f, keys)
        writer.writeheader()
        writer.writerows(parameterSpace)
    """



"""
Character width dictionary and convenience functions for column sizing
with xlwt when Arial 10 is the standard font.  Widths were determined
experimentally using Excel 2000 on Windows XP.  I have no idea how well
these will work on other setups.  For example, I don't know if system
video settings will affect the results.  I do know for sure that this
module won't be applicable to other fonts in general.
//John Yeung  2009-09-02
"""

charwidths = {
    "0": 262.637,
    "1": 262.637,
    "2": 262.637,
    "3": 262.637,
    "4": 262.637,
    "5": 262.637,
    "6": 262.637,
    "7": 262.637,
    "8": 262.637,
    "9": 262.637,
    "a": 262.637,
    "b": 262.637,
    "c": 262.637,
    "d": 262.637,
    "e": 262.637,
    "f": 146.015,
    "g": 262.637,
    "h": 262.637,
    "i": 117.096,
    "j": 88.178,
    "k": 233.244,
    "l": 88.178,
    "m": 379.259,
    "n": 262.637,
    "o": 262.637,
    "p": 262.637,
    "q": 262.637,
    "r": 175.407,
    "s": 233.244,
    "t": 117.096,
    "u": 262.637,
    "v": 203.852,
    "w": 321.422,
    "x": 203.852,
    "y": 262.637,
    "z": 233.244,
    "A": 321.422,
    "B": 321.422,
    "C": 350.341,
    "D": 350.341,
    "E": 321.422,
    "F": 291.556,
    "G": 350.341,
    "H": 321.422,
    "I": 146.015,
    "J": 262.637,
    "K": 321.422,
    "L": 262.637,
    "M": 379.259,
    "N": 321.422,
    "O": 350.341,
    "P": 321.422,
    "Q": 350.341,
    "R": 321.422,
    "S": 321.422,
    "T": 262.637,
    "U": 321.422,
    "V": 321.422,
    "W": 496.356,
    "X": 321.422,
    "Y": 321.422,
    "Z": 262.637,
    " ": 146.015,
    "!": 146.015,
    '"': 175.407,
    "#": 262.637,
    "$": 262.637,
    "%": 438.044,
    "&": 321.422,
    "'": 88.178,
    "(": 175.407,
    ")": 175.407,
    "*": 203.852,
    "+": 291.556,
    ",": 146.015,
    "-": 175.407,
    ".": 146.015,
    "/": 146.015,
    ":": 146.015,
    ";": 146.015,
    "<": 291.556,
    "=": 291.556,
    ">": 291.556,
    "?": 262.637,
    "@": 496.356,
    "[": 146.015,
    "\\": 146.015,
    "]": 146.015,
    "^": 203.852,
    "_": 262.637,
    "`": 175.407,
    "{": 175.407,
    "|": 146.015,
    "}": 175.407,
    "~": 291.556,
}

# By default, Excel displays column widths in units equal to the width
# of '0' (the zero character) in the standard font.  For me, this is
# Arial 10, but it can be changed by the user.  The BIFF file format
# stores widths in units 1/256th that size.
#
# Within Excel, the smallest incrementable amount for column width
# is the pixel.  However many pixels it takes to draw '0' is how many
# increments there are between a width of 1 and a width of 2.  A
# request for a finer increment will be rounded to the nearest pixel.
# For Arial 10, this is 9 pixels, but different fonts will of course
# require different numbers of pixels, and thus have different column
# width granularity.
#
# So far so good, but there is a wrinkle.  Excel pads the first unit
# of column width by 7 pixels.  At least this is the padding when the
# standard font is Arial 10 or Courier New 10, the two fonts I've tried.
# It don't know if it's different for different fonts.  For Arial 10,
# with a padding of 7 pixels and a 9-pixel-wide '0', this results in 16
# increments to get from width 0 (hidden) to width 1.  Ten columns of
# width 1 are 160 pixels wide while five columns of width 2 are 125
# pixels wide.  A single column of width 10 is only 97 pixels wide.
#
# The punch line is that pixels are the true measure of width, and
# what Excel reports as the column width is wonky between 0 and 1.
# The only way I know to find out the padding for a desired font is
# to set that font as the standard font in Excel and count pixels.


def colwidth(n):
    """Translate human-readable units to BIFF column width units"""
    if n <= 0:
        return 0
    if n <= 1:
        return n * 456
    return 200 + n * 256


def fitwidth(data, bold=False):
    """Try to autofit Arial 10"""
    maxunits = 0
    for ndata in data.split("\n"):
        units = 220
        for char in ndata:
            if char in charwidths:
                units += charwidths[char]
            else:
                units += charwidths["0"]
        if maxunits < units:
            maxunits = units
    if bold:
        maxunits *= 1.1
    return max(maxunits, 700)  # Don't go smaller than a reported width of 2


def fitheight(data, bold=False):
    """Try to autofit Arial 10"""
    rowlen = len(data.split("\n"))
    if rowlen > 1:
        units = 230 * rowlen
    else:
        units = 290
    if bold:
        units *= 1.1
    return int(units)



from bionetgen.core.utils.logging import BNGLogger

logger = BNGLogger()

# All classes that deal with patterns
class Pattern:
    """
    Pattern object. Fundamentally it's just a list of molecules
    which are defined later.

    Attributes
    ----------
    _bonds : Bonds
        setting a pattern requires you to keep track of all bonds to
        correctly label them, this object tracks everything
    compartment : str
        compartment of the overall pattern (not the same thing as
        molecule compartment, those have their own)
    _label : str
        label of the overall pattern (not the same thing as molecule
        label, those have their own)
    molecules : list[Molecule]
        list of molecule objects that are in the pattern
    fixed : bool
        used for constant species, sets "$" at the beginning of the
        pattern string
    MatchOnce : bool
        used for matchOnce syntax, "{MatchOnce}PatternStr"
    """

    def __init__(
        self, molecules=[], bonds=None, compartment=None, label=None, canonicalize=False
    ):
        self.molecules = molecules
        self._bonds = bonds
        self.compartment = compartment
        self.label = label
        self.fixed = False
        self.MatchOnce = False
        self.relation = None
        self.quantity = None
        self.nautyG = None
        self.canonical_certificate = None
        self.canonical_label = None
        if canonicalize:
            self.canonicalize()

    def canonicalize(self):
        # set a location for logging
        loc = f"{__file__} : Pattern.canonicalize()"
        # try importing pynauty to canonicalize the labeling
        try:
            import pynauty
        except ImportError:
            logger.warning(
                f"Importing pynauty failed, cannot canonicalize. Pattern equality checking is not guaranteed to work for highly symmetrical species.",
                loc=loc,
            )
            return
        # find how many vertices we need
        lmol = len(self.molecules)
        lcomp = sum([len(x.components) for x in self.molecules])
        node_cnt = lmol + lcomp
        # initialize our pynauty graph
        G = pynauty.Graph(node_cnt)
        # going to need to figure out bonding
        bond_dict = {}
        # save our IDs
        rev_grpIds = {}
        grpIds = {}
        # also pointers to each object
        node_ptrs = {}
        bond_node_ptrs = {}
        # we'll need to seutp coloring
        colors = {}
        currId = 0
        mCopyId = 0
        cCopyId = 0
        # let's loop over everything in the pattern
        for molec in self.molecules:
            # setting colors
            color_id = (molec.name, None, None)
            if color_id in colors:
                colors[color_id].add(currId)
            else:
                colors[color_id] = set([currId])
            # saving IDs
            parent_id = (molec.name, None, mCopyId, cCopyId)
            if parent_id in grpIds:
                mCopyId += 1
                parent_id = (molec.name, None, mCopyId, cCopyId)
                grpIds[parent_id] = currId
            else:
                grpIds[parent_id] = currId
            rev_grpIds[currId] = parent_id
            node_ptrs[currId] = molec
            currId += 1
            # now looping over components
            for comp in molec.components:
                # saving component coloring
                comp_color_id = (molec.name, comp.name, comp.state)
                if comp_color_id in colors:
                    colors[comp_color_id].add(currId)
                else:
                    colors[comp_color_id] = set([currId])
                chid_id = (molec.name, comp.name, mCopyId, cCopyId)
                # connecting the component to the molecule
                G.connect_vertex(grpIds[parent_id], [currId])
                # saving component IDs
                if chid_id in grpIds:
                    cCopyId += 1
                    chid_id = (molec.name, comp.name, mCopyId, cCopyId)
                    grpIds[chid_id] = currId
                else:
                    grpIds[chid_id] = currId
                rev_grpIds[currId] = chid_id
                node_ptrs[currId] = comp
                currId += 1
                # saving bonds
                if len(comp._bonds) != 0:
                    for bond in comp._bonds:
                        if bond not in bond_dict.keys():
                            bond_dict[bond] = [chid_id]
                        else:
                            bond_dict[bond].append(chid_id)
        # now we got everything, we implement it in the graph
        for bond in bond_dict:
            # check if each of our bonds have exactly two end points
            if len(bond_dict[bond]) == 2:
                id1 = bond_dict[bond][0]
                id1 = grpIds[id1]
                id2 = bond_dict[bond][1]
                id2 = grpIds[id2]
                G.connect_vertex(id1, [id2])
            else:
                # raise a warning
                logger.warning(
                    f"Bond {bond} doesn't have exactly 2 end points, please check that you don't have any dangling bonds.",
                    loc=loc,
                )
        # we get our color sets
        color_sets = list(colors.values())
        # set vertex coloring
        G.set_vertex_coloring(color_sets)
        # save our graph
        self.nautyG = G
        # generate the canonical certificate for the entire graph
        self.canonical_certificate = pynauty.certificate(self.nautyG)
        # generate the canonical label for the entire graph
        # first, we give every node their canonical order
        canon_order = pynauty.canon_label(self.nautyG)
        for iordr, ordr in enumerate(canon_order):
            node_ptrs[ordr].canonical_order = iordr
        # relabeling bonds
        relabeling_bond_dict = {}
        for bond in bond_dict:
            # check if each of our bonds have exactly two end points
            if len(bond_dict[bond]) == 2:
                id1 = bond_dict[bond][0]
                id1 = grpIds[id1]
                comp1 = node_ptrs[id1]
                id2 = bond_dict[bond][1]
                id2 = grpIds[id2]
                comp2 = node_ptrs[id2]
                parent_order = min(
                    comp1.parent_molecule.canonical_order,
                    comp2.parent_molecule.canonical_order,
                )
                comp_order = min(comp1.canonical_order, comp2.canonical_order)
                relabeling_bond_dict[(parent_order, comp_order)] = (comp1, comp2)
            else:
                # raise a warning
                logger.warning(
                    f"Bond {bond} doesn't have exactly 2 end points, please check that you don't have any dangling bonds.",
                    loc=loc,
                )
        # this will give us the keys to canonically sorted bonds
        sorted_order = sorted(relabeling_bond_dict.keys())
        for ibond, sbond in enumerate(sorted_order):
            # now we add a canonical bond ID to each component
            c1, c2 = relabeling_bond_dict[sbond]
            if c1.canonical_bonds is None:
                c1.canonical_bonds = [str(ibond + 1)]
            else:
                c1.canonical_bonds.append(str(ibond + 1))
            if c2.canonical_bonds is None:
                c2.canonical_bonds = [str(ibond + 1)]
            else:
                c2.canonical_bonds.append(str(ibond + 1))
        # and now we can get the canonical label
        self.canonical_label = self.print_canonical()

    def print_canonical(self):
        # need to make sure we don't print useless compartments
        self.consolidate_molecule_compartments()
        canon_label = ""
        # we first deal with the pattern compartment
        if self.compartment is not None:
            canon_label += "@{}".format(self.compartment)
        if self.label is not None:
            canon_label += "%{}".format(self.label)
        if self.label is not None or self.compartment is not None:
            canon_label += ":"
        # now loop over all molecules in canonical order
        canon_ords = [m.canonical_order for m in self.molecules]
        canon_ord_pairs = zip(range(len(self.molecules)), canon_ords)
        sorted_canon_ord_pairs = sorted(canon_ord_pairs, key=lambda x: x[1])
        for imol, mol in enumerate(sorted_canon_ord_pairs):
            mol_id = mol[0]
            if imol == 0:
                if self.fixed:
                    canon_label += "$"
                if self.MatchOnce:
                    canon_label += "{MatchOnce}"
            if imol > 0:
                canon_label += "."
            canon_label += self.molecules[mol_id].print_canonical()
        if self.relation is not None:
            canon_label += f"{self.relation}{self.quantity}"
        return canon_label

    def __contains__(self, val):
        return val in self.molecules

    def __eq__(self, other):
        loc = f"{__file__} : Pattern.__eq__()"
        if isinstance(other, Pattern):
            logger.debug(f"Comparison class matches: {other.__class__}", loc=loc)
            # checking pattern-wide properties
            if (other.compartment == self.compartment) and (other.label == self.label):
                logger.debug(
                    f"Compartment or label matches: {other.compartment}, {other.label}",
                    loc=loc,
                )
                # checking mods
                if (other.fixed == self.fixed) and (other.MatchOnce == self.MatchOnce):
                    logger.debug(
                        f"fixed or matchonce matches: {other.fixed}, {other.MatchOnce}",
                        loc=loc,
                    )
                    # checking quantifiers
                    if (other.relation == self.relation) and (
                        other.quantity == self.quantity
                    ):
                        logger.debug(
                            f"relation or quantity matches: {other.relation}, {other.quantity}",
                            loc=loc,
                        )
                        # if we made the label, we can just compare the two
                        if (self.canonical_label is not None) and (
                            other.canonical_label is not None
                        ):
                            return self.canonical_label == other.canonical_label
                        # now we can check contents
                        for molecule in self:
                            if molecule not in other.molecules:
                                logger.debug(
                                    f"molecule doesn't match: {molecule}", loc=loc
                                )
                                return False
                        # isomorphism check if we have the certificate
                        if (self.canonical_certificate is not None) and (
                            other.canonical_certificate is not None
                        ):
                            if (
                                self.canonical_certificate
                                != other.canonical_certificate
                            ):
                                return False
                        # TODO: molecules match, check bonds
                        # Bonds match, patterns are the same
                        logger.debug("patterns match!", loc=loc)
                        return True
        return False

    @property
    def compartment(self):
        return self._compartment

    @compartment.setter
    def compartment(self, value):
        # TODO: Build in logic to set the
        # outer compartment
        # print("Warning: Logical checks are not complete")
        self._compartment = value

    def consolidate_molecule_compartments(self):
        # if the molecule compartment matches overall pattern
        # compartment, don't print the molecule compartments
        overall_comp = self.compartment
        if overall_comp is not None:
            for molec in self.molecules:
                if molec.compartment == overall_comp:
                    molec.compartment = None

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # TODO: Build in logic to set
        # the outer label
        # print("Warning: Logical checks are not complete")
        self._label = value

    def __str__(self):
        # need to make sure we don't print useless compartments
        self.consolidate_molecule_compartments()
        sstr = ""
        # we first deal with the pattern compartment
        if self.compartment is not None:
            sstr += "@{}".format(self.compartment)
        if self.label is not None:
            sstr += "%{}".format(self.label)
        if self.label is not None or self.compartment is not None:
            sstr += ":"
        # now loop over all molecules
        for imol, mol in enumerate(self.molecules):
            if imol == 0:
                if self.fixed:
                    sstr += "$"
                if self.MatchOnce:
                    sstr += "{MatchOnce}"
            if imol > 0:
                sstr += "."
            sstr += str(mol)
        if self.relation is not None:
            sstr += f"{self.relation}{self.quantity}"
        return sstr

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return self.molecules[key]

    def __iter__(self):
        return self.molecules.__iter__()

    # TODO: Implement __contains__


class Molecule:
    """
    Molecule object. A pattern is a list of molecules.
    This object also handles molecule types where components
    have a list of possible states.

    Attributes
    ----------
    _name : str
        name of the molecule
    _compartment : str
        compartment of the molecule
    _label : str
        label of the molecule
    _components : list[Component]
        list of components for this molecule

    Methods
    -------
    add_component(name, state=None, states=[])
        add a component object to the list of components with name
        "name", current state "state" or a list of states
        (for molecule types) "states"
    """

    def __init__(self, name="0", components=[], compartment=None, label=None):
        self._name = name
        self._components = components
        self._compartment = compartment
        self._label = label
        self.canonical_order = None
        self.canonical_label = None
        self.parent_pattern = None

    def __contains__(self, val):
        return val in self.components

    def __eq__(self, other):
        loc = f"{__file__} : Molecule.__eq__()"
        # check object type
        if isinstance(other, Molecule):
            logger.debug(f"Comparison class matches: {other.__class__}", loc=loc)
            # check attributes
            if (
                (other.name == self.name)
                and (other.compartment == self.compartment)
                and (other.label == self.label)
            ):
                logger.debug(
                    f"name, compartment and labels match: {other.name}, {other.compartment}, {other.label}",
                    loc=loc,
                )
                if (self.canonical_label is not None) and (
                    other.canonical_label is not None
                ):
                    # we can check canonical labels
                    if self.canonical_label != other.canonical_label:
                        return False
                # check components now
                for component in self:
                    if component not in other.components:
                        logger.debug(f"component doesn't match: {component}", loc=loc)
                        return False
                # everything matches
                logger.debug("molecules match", loc=loc)
                return True
        return False

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.components[key]

    def __iter__(self):
        return self.components.__iter__()

    # TODO: implement __setitem__,  __contains__

    def __str__(self):
        mol_str = self.name
        # we have a null species
        if not self.name == "0":
            mol_str += "("
        # we _could_ just not do () if components
        # don't exist but that has other issues,
        # especially for extension highlighting
        if len(self.components) > 0:
            for icomp, comp in enumerate(self.components):
                if icomp > 0:
                    mol_str += ","
                mol_str += str(comp)
        # we have a null species
        if not self.name == "0":
            mol_str += ")"
        if self.compartment is not None:
            mol_str += "@{}".format(self.compartment)
        if self.label is not None:
            mol_str += "%{}".format(self.label)
        return mol_str

    def print_canonical(self):
        # print in canonical order
        canon_label = self.name
        # we have a null species
        if not self.name == "0":
            canon_label += "("
        # we _could_ just not do () if components
        # don't exist but that has other issues,
        # especially for extension highlighting
        if len(self.components) > 0:
            canon_ords = [c.canonical_order for c in self.components]
            canon_ord_pairs = zip(range(len(self.components)), canon_ords)
            sorted_canon_ord_pairs = sorted(canon_ord_pairs, key=lambda x: x[1])
            for icomp, comp in enumerate(sorted_canon_ord_pairs):
                comp_id = comp[0]
                if icomp > 0:
                    canon_label += ","
                canon_label += self.components[comp_id].print_canonical()
        # we have a null species
        if not self.name == "0":
            canon_label += ")"
        if self.compartment is not None:
            canon_label += "@{}".format(self.compartment)
        if self.label is not None:
            canon_label += "%{}".format(self.label)
        return canon_label

    ### PROPERTIES ###
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        # print("Warning: Logical checks are not complete")
        # TODO: Check for invalid characters
        self._name = value

    @property
    def components(self):
        return self._components

    @components.setter
    def components(self, value):
        # print("Warning: Logical checks are not complete")
        self._components = value

    def __repr__(self):
        return str(self)

    @property
    def compartment(self):
        return self._compartment

    @compartment.setter
    def compartment(self, value):
        # print("Warning: Logical checks are not complete")
        self._compartment = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # print("Warning: Logical checks are not complete")
        self._label = value

    def _add_component(self, name, state=None, states=[]):
        comp_obj = Component()
        comp_obj.name = name
        comp_obj.state = state
        comp_obj.states = states
        self.components.append(comp_obj)

    def add_component(self, name, state=None, states=[]):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._add_component(name, state, states)


class Component:
    """
    Component object that describes the state, label and bonding
    for each component. Molecules can optionally contain components

    Attributes
    ----------
    name : str
        name of the component
    _label : str
        label of the component
    _state : str
        state of the component, not used for molecule types
    _states : list[str]
        list of states for molecule types
    _bonds : list[Bond]
        list of bond objects that describes bonding of the component

    Methods
    -------
    add_state()
        not implemented. will eventually be used to add additional states
        to an existing component
    add_bond()
        not implemented. will eventually be used to add additional bonds
        to an existing component
    """

    def __init__(self):
        self._name = ""
        self._label = None
        self._state = None
        self._states = []
        self._bonds = []
        self.canonical_label = None
        self.canonical_order = None
        self.canonical_bonds = None
        self.parent_molecule = None

    def __eq__(self, other):
        loc = f"{__file__} : Component.__eq__()"
        # check type
        # import ipdb;ipdb.set_trace()
        if isinstance(other, Component):
            logger.debug(f"Comparison class matches: {other.__class__}", loc=loc)
            # check attributes
            if (other.name == self.name) and (other.label == self.label):
                logger.debug(
                    f"name and labels match: {other.name}, {other.label}", loc=loc
                )
                # check states
                if len(other.states) == len(self.states):
                    logger.debug(f"state lists match: {other.states}", loc=loc)
                    # check current state
                    if other.state == self.state:
                        logger.debug(f"states match: {other.state}", loc=loc)
                        if (self.canonical_label is not None) and (
                            other.canonical_label is not None
                        ):
                            # we can check canonical labels
                            if self.canonical_label != other.canonical_label:
                                return False
                        # check bonds
                        # TODO: try to decide if A(b!1).B(a!1) is the same
                        # as A(b!2).B(a!2), if so, the bond check is much harder
                        # for bond in self.bonds:
                        #     if bond not in other.bonds:
                        #         logger.debug(
                        #             f"bonds don't match!: {other.bonds}", loc=loc
                        #         )
                        #         return False
                        if len(self.bonds) == len(other.bonds):
                            logger.debug("components match", loc=loc)
                            return True
        return False

    def __repr__(self):
        return str(self)

    def __str__(self):
        comp_str = self.name
        # only for molecule types
        if len(self.states) > 0:
            for istate, state in enumerate(self.states):
                comp_str += "~{}".format(state)
        # for any other pattern
        if self.state is not None:
            comp_str += "~{}".format(self.state)
        if self.label is not None:
            comp_str += "%{}".format(self.label)
        if len(self.bonds) > 0:
            for bond in self.bonds:
                comp_str += "!{}".format(bond)
        return comp_str

    def print_canonical(self):
        comp_str = self.name
        # only for molecule types
        if len(self.states) > 0:
            for istate, state in enumerate(self.states):
                comp_str += "~{}".format(state)
        # for any other pattern
        if self.state is not None:
            comp_str += "~{}".format(self.state)
        if self.label is not None:
            comp_str += "%{}".format(self.label)
        if self.canonical_bonds is not None:
            for bond in self.canonical_bonds:
                comp_str += "!{}".format(bond)
        return comp_str

    ### PROPERTIES ###
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._name = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._label = value

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._state = value

    @property
    def states(self):
        return self._states

    @states.setter
    def states(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._states = value

    @property
    def bonds(self):
        return self._bonds

    @bonds.setter
    def bonds(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._bonds = value

    def _add_state(self):
        raise NotImplementedError

    def add_state(self):
        self._add_state()

    def _add_bond(self):
        raise NotImplementedError

    def add_bond(self):
        self._add_bond()



import platform, os, subprocess
from cement import init_defaults
from cement.utils.version import get_version_banner
from .version import get_version


def get_latest_bng_version():
    """
    Pulls the currently used BNG version from assets/BNGVERSION file
    """
    # get the version tag for the BNG version used
    libpath = os.path.abspath(__file__)
    libpath = libpath.split(os.path.sep)
    verpath = os.path.sep.join(libpath[:-2] + ["assets", "BNGVERSION"])
    if os.path.isfile(verpath):
        with open(verpath) as f:
            tag = f.read()
        return tag
    else:
        return "UNKNOWN"


class BNGDefaults:
    def __init__(self):
        """
        A class to define the default configuration for cement apps

        Attributes
        ----------
        system : str
            the name of the OS that's running the app
        bng_name : str
            OS appropriate name of the BNG folder
        bng_path : str
            full absolute path to the BNG folder
        lib_path : str
            path to CLI library
        stdout : str
            the name of the subprocess attribute to pass stdout to
        stderr : str
            the name of the subprocess attribute to pass stderr to
        config : dict
            dictionary containing the application defaults
        banner : str
            app banner that gets printed when ran with -v
        """
        # determine what bng we are using
        system = platform.system()
        if system == "Linux":
            bng_name = "bng-linux"
        elif system == "Windows":
            bng_name = "bng-win"
        elif system == "Darwin":
            bng_name = "bng-mac"
        # set attributes
        self.system = system
        self.bng_name = bng_name
        # configuration defaults
        CONFIG = init_defaults("bionetgen")
        lib_path = os.path.dirname(__file__)
        lib_path = os.path.split(lib_path)[0]
        CONFIG["bionetgen"]["bngpath"] = os.path.join(lib_path, bng_name)
        # notebook
        CONFIG["bionetgen"]["notebook"] = {}
        CONFIG["bionetgen"]["notebook"]["path"] = os.path.join(
            lib_path, "assets", "bionetgen.ipynb"
        )
        CONFIG["bionetgen"]["notebook"]["template"] = os.path.join(
            lib_path, "assets", "bionetgen-temp.ipynb"
        )
        CONFIG["bionetgen"]["notebook"]["name"] = "bng-notebook.ipynb"
        # cvode paths
        CONFIG["bionetgen"]["cvode_lib"] = None
        CONFIG["bionetgen"]["cvode_include"] = None
        # set attributes
        self.bng_path = os.path.join(lib_path, bng_name)
        self.lib_path = lib_path
        # version banner
        VERSION_BANNER = """BioNetGen simple command line interface {}\nBioNetGen version: {}\n{}
        """.format(
            get_version(), get_latest_bng_version(), get_version_banner()
        )
        # set attributes
        self.banner = VERSION_BANNER
        # stdout
        CONFIG["bionetgen"]["stdout"] = "PIPE"
        CONFIG["bionetgen"]["stderr"] = "STDOUT"
        self.stdout = subprocess.PIPE
        self.stderr = subprocess.PIPE
        self.config = CONFIG


defaults = BNGDefaults()



from bionetgen.atomizer.utils import readBNGXML
import argparse
from collections import defaultdict

# from cStringIO import StringIO
from io import StringIO
from bionetgen.atomizer.utils import extractAtomic
from copy import deepcopy
import networkx as nx


def molecule2stateTuples(molecule):
    """
    Receives a molecule structure, returns a tuple detailing the state of the contained states
    """
    tupleList = []
    for component in molecule.components:
        componentDefinition = [component.name]
        componentDefinition.append(1 if len(component.bonds) > 0 else 0)
        componentDefinition.append(
            "" if len(component.states) == 0 else component.activeState
        )
        tupleList.append(tuple(componentDefinition))
    return tupleList


def extractCenterContext(rules, excludeReverse=False):
    transformationCenter = []
    transformationContext = []
    transformationProduct = []
    atomicArray = []
    actionNames = []
    label = []
    doubleModificationRules = []
    for idx, rule in enumerate(rules):
        (
            tatomicArray,
            ttransformationCenter,
            ttransformationContext,
            tproductElements,
            tactionNames,
            tlabelArray,
        ) = extractAtomic.extractTransformations([rule], True)
        if (
            excludeReverse
            and "_reverse_" in rule[0].label
            and len([x for x in tactionNames if "ChangeCompartment" not in x]) > 1
        ):
            continue
        label.append(rule[0].label)

        if len([x for x in tactionNames if "ChangeCompartment" not in x]) > 1:
            doubleModificationRules.append(rule[0].label)
        transformationCenter.append(ttransformationCenter)
        transformationContext.append(ttransformationContext)

        actionNames.append(tactionNames)
        atomicArray.append(tatomicArray)
        transformationProduct.append(tproductElements)
    return (
        label,
        transformationCenter,
        transformationContext,
        transformationProduct,
        atomicArray,
        actionNames,
        doubleModificationRules,
    )


def askQuestions(inputfile, molecule, center, context=None):
    _, rules, _ = readBNGXML.parseXML(inputfile)

    ruleArray = []
    contextArray = []
    for idx, rule in enumerate(rules):
        (
            tatomicArray,
            ttransformationCenter,
            ttransformationContext,
            tproductElements,
            tactionNames,
            tlabelArray,
        ) = extractAtomic.extractTransformations([rule], True)

        if any(
            [molecule in y and center in y for x in ttransformationCenter for y in x]
        ):
            if context:
                ruleArray.append(str(rule[0]).split(":")[0])
                contextArray.append(
                    [
                        y
                        for x in ttransformationContext
                        for y in x
                        if context in y and molecule in y
                    ]
                )
            else:
                print(rule)
    return ruleArray, contextArray


def getChemicalStates(rules):
    """
    goes through a list of rules and creates a dictionary (molecules) of possible chemical states of its components
    """
    chemicalStates = defaultdict(list)
    for rule in rules:
        for chemicalSet in [rule[0].reactants, rule[0].products]:
            for chemical in chemicalSet:
                for molecule in chemical.molecules:
                    stateTuple = molecule2stateTuples(molecule)
                    chemicalStates[molecule.name].append(stateTuple)

    return chemicalStates


def patternsFromString(chemical):
    patterndict = defaultdict(list)
    chemicalString = str(chemical)
    molecules = chemicalString.split(".")
    for molecule in molecules:
        partitions = molecule.split("(")
        if "!" in partitions[1]:
            componentName = partitions[1].split("!")[0]
            patterndict[partitions[0]].append((componentName, 1, ""))
        elif "~" in partitions[1]:
            componentName = partitions[1].split("~")[0]
            stateName = partitions[1].split("~")[1][:-1]
            patterndict[partitions[0]].append((componentName, 0, stateName))
        else:
            componentName = partitions[1][:-1]
            patterndict[partitions[0]].append((componentName, 0, ""))
    return patterndict


def getRestrictedChemicalStates(labelArray, products, contexts, doubleAction):
    """
    goes through a list of reaction center and their associated context and creates a dictionary of molecules
    of possible chemical states
    """

    # sortedChemicalStates = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
    sortedChemicalStates = defaultdict(
        lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    )
    doubleActionDict = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        )
    )
    for label, product, context in zip(labelArray, products, contexts):
        # reaction centers are also the context for other reaction centers. We deal with them separately
        if label in doubleAction:
            pDict = defaultdict(list)
            for referenceProduct in product:
                for p in referenceProduct:
                    result = patternsFromString(p)
                    for molecule in result:
                        for pattern in result[molecule]:
                            pDict[molecule].append(pattern)
            pDict2 = deepcopy(pDict)
            for molecule in pDict:
                for componentState in pDict[molecule]:
                    for componentState2 in [
                        x for x in pDict2[molecule] if x[0] != componentState[0]
                    ]:
                        isActive1 = componentState[1] == 1 or componentState[2] not in [
                            "",
                            "0",
                        ]
                        isActive2 = componentState2[1] == 1 or componentState2[
                            2
                        ] not in ["", "0"]
                        moleculeName = molecule.split("%")[0]
                        doubleActionDict[moleculeName][componentState[0]][isActive1][
                            componentState2[0]
                        ][isActive2].append(label)

        for indvproduct, indvcontext in zip(product, context):
            pDict = defaultdict(list)
            cDict = defaultdict(list)
            for p in indvproduct:
                result = patternsFromString(p)
                for molecule in result:
                    for pattern in result[molecule]:
                        pDict[molecule].append(pattern)
            for c in indvcontext:
                result = patternsFromString(c)
                for molecule in result:
                    for pattern in result[molecule]:
                        cDict[molecule].append(pattern)
            for molecule in pDict:
                for componentState in pDict[molecule]:

                    # FIXME: This is to account for dimers where or places where there is more than one components with the same name. Truly this should be enother kind of classification
                    for componentState2 in [
                        x for x in cDict[molecule] if x[0] != componentState[0]
                    ]:
                        sortedChemicalStates[molecule][componentState][
                            componentState2[0]
                        ][componentState2[1:]].append(label)
                        # sortedChemicalStates[molecule][componentState][componentState2[0]].add(componentState2[1:])
    return sortedChemicalStates, doubleActionDict


def sortChemicalStates(chemicalStates):
    """
    sorts a dictionary of chemicalStates into a dictionary of molecules of chemicalStates
    of their associated-same-molecule-components (and their states)
    """
    sortedChemicalStates = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
    for molecule in chemicalStates:
        for stateInstance in chemicalStates[molecule]:
            for componentState in stateInstance:
                for componentState2 in stateInstance:
                    if componentState == componentState2:
                        continue
                    sortedChemicalStates[molecule][componentState][
                        componentState2[0]
                    ].add(componentState2[1:])
    return sortedChemicalStates


def isActive(state):
    if state[0] == 1 or state[1] not in ["", "0"]:
        return True
    return False


def componentStateSize(molecules, moleculeName, componentName):
    for molecule in molecules:
        if molecule.name == moleculeName:
            for component in molecule.components:
                if component.name == componentName:
                    return max(2, len(component.states))


def analyzeDependencies(
    componentStateCollection, state, moleculeName, molecules, dependencies
):
    for componentName in componentStateCollection:
        stateSize = componentStateSize(molecules, moleculeName, componentName)

        if stateSize == len(componentStateCollection[componentName]):
            dependencies[moleculeName]["independent"].add((state, componentName))
            # print(moleculeName,state,componentName,componentStateCollection[componentName])
        elif len(componentStateCollection[componentName]) == 1:
            activeState = list(componentStateCollection[componentName])[0]
            if isActive((state[1], state[2])) and isActive(activeState):
                dependencies[moleculeName]["requirement"].add(
                    ((componentName, activeState[0], activeState[1]), state)
                )
            elif isActive((state[1], state[2])) and not isActive(activeState):
                dependencies[moleculeName]["nullrequirement"].add(
                    (((componentName, activeState[0], activeState[1])), state)
                )

            # elif (not isActive((state[1], state[2]))) and isActive(activeState):
            #    print(moleculeName, componentName,activeState,state)
            # dependencies[moleculeName]['nullrequirement'].add(((state, (componentName, activeState[0], activeState[1]))))


def detectDependencies(stateDictionary, molecules):
    dependencies = defaultdict(lambda: defaultdict(set))
    preprocessing = defaultdict(lambda: defaultdict(dict))
    # preprocess for dimer information
    for moleculeName in stateDictionary:
        parsedMoleculeName = moleculeName.split("%")[0]
        # parsedMoleculeName = moleculeName
        for state in stateDictionary[moleculeName]:
            preprocessing[parsedMoleculeName][state].update(
                stateDictionary[moleculeName][state]
            )
    for moleculeName in preprocessing:
        for state in preprocessing[moleculeName]:
            analyzeDependencies(
                preprocessing[moleculeName][state],
                state,
                moleculeName,
                molecules,
                dependencies,
            )
    return dependencies


def reverseState(moleculeName, state, molecules):
    for molecule in molecules:
        if molecule.name == moleculeName:
            for component in molecule.components:
                if component.name == state[0]:
                    if len(component.states) <= 1:
                        if isActive(state[1:]):
                            return (state[0], 0, "")
                        else:
                            return (state[0], 1, "")
                    else:
                        if isActive(state[1:]):
                            return (state[0], 0, "0")
                        else:
                            return (
                                state[0],
                                0,
                                [x for x in component.states if x != "0"][0],
                            )


def getMutualExclusions(stateDictionary, molecules):
    for molecule in stateDictionary:
        stateDictionary[molecule]["exclusion"] = [
            x
            for x in stateDictionary[molecule]["nullrequirement"]
            if (
                reverseState(molecule, x[1], molecules),
                reverseState(molecule, x[0], molecules),
            )
            in stateDictionary[molecule]["nullrequirement"]
        ]
        stateDictionary[molecule]["nullrequirement"] = [
            x
            for x in stateDictionary[molecule]["nullrequirement"]
            if x not in stateDictionary[molecule]["exclusion"]
        ]


def getMotifRelationships(stateDictionary, molecules):
    motifDictionary = defaultdict(lambda: defaultdict(list))
    for molecule in stateDictionary:
        motifDictionary[molecule]["exclusion"] = set(
            [
                tuple(sorted([x[0][0], x[1][0]]))
                for x in stateDictionary[molecule]["nullrequirement"]
                if (
                    reverseState(molecule, x[1], molecules),
                    reverseState(molecule, x[0], molecules),
                )
                in stateDictionary[molecule]["nullrequirement"]
            ]
        )
        motifDictionary[molecule]["inclusion"] = set(
            [
                tuple(sorted([x[0][0], x[1][0]]))
                for x in stateDictionary[molecule]["requirement"]
                if (x[1], x[0]) in stateDictionary[molecule]["requirement"]
            ]
        )

        motifDictionary[molecule]["ordering"] = set(
            [
                tuple([x[0][0], x[1][0]])
                for x in stateDictionary[molecule]["requirement"]
                if (reverseState(molecule, x[1], molecules), x[0])
                in stateDictionary[molecule]["nullrequirement"]
            ]
        )
        motifDictionary[molecule]["fullIndependence"] = set(
            [
                tuple(sorted([x[0][0], x[1]]))
                for x in stateDictionary[molecule]["independent"]
                if isActive(x[0][1:])
                and any(
                    x[1] == y[0][0] and x[0][0] == y[1] and isActive(y[0][1:])
                    for y in stateDictionary[molecule]["independent"]
                )
            ]
        )

        motifDictionary[molecule]["partialIndependence+"] = set(
            [
                tuple([x[0][0], x[1]])
                for x in stateDictionary[molecule]["independent"]
                if isActive(x[0][1:])
                and any(
                    x[1] in y[1] and x[0][0] in y[0]
                    for y in stateDictionary[molecule]["requirement"]
                )
            ]
        )

        motifDictionary[molecule]["partialIndependence-"] = set(
            [
                tuple([x[0][0], x[1]])
                for x in stateDictionary[molecule]["independent"]
                if isActive(x[0][1:])
                and any(
                    x[1] in y[1] and x[0][0] in y[0]
                    for y in stateDictionary[molecule]["nullrequirement"]
                )
            ]
        )
        motifDictionary[molecule]["repression"] = set(
            [tuple([x[0], x[1]]) for x in stateDictionary[molecule]["repression"]]
        )
    return motifDictionary


def removeIndirectDependencies(dependencies, stateSpace, motifFlag=False):
    """
    goes through the list of dependencies and search for those states that requires more than 1 condition.
    If true, it will check if any of the set of prerequirements is a sufficient conditions for any of the other prerequirements
    to be valid. If true it will remove those redundant prerequirements them
    """

    relationship = "ordering" if motifFlag else "requirement"
    indirect = defaultdict(set)
    for molecule in dependencies:

        for requirement in set([x[1] for x in dependencies[molecule][relationship]]):
            prerequirements = [
                x[0]
                for x in dependencies[molecule][relationship]
                if requirement == x[1]
            ]
            # if a particular state has more than one requirement
            if len(prerequirements) > 1:
                candidateMolecules = [x for x in stateSpace.keys() if molecule in x]
                relevantStateSpace = defaultdict(lambda: defaultdict(set))
                for candidate in candidateMolecules:
                    for x in prerequirements:
                        if motifFlag:
                            for referenceCandidate in stateSpace[candidate]:
                                if x in referenceCandidate and isActive(
                                    referenceCandidate
                                ):
                                    relevantStateSpace[x].update(
                                        stateSpace[candidate][referenceCandidate]
                                    )
                        else:
                            relevantStateSpace[x].update(
                                stateSpace[candidate][x]
                            )  # = {x: stateSpace[molecule][x] for x in prerequirements}

                for rsp in relevantStateSpace:
                    for partner in relevantStateSpace[rsp]:
                        partnerDefinition = list(relevantStateSpace[rsp][partner])
                        if motifFlag:
                            # is any single prerequirement a sufficient condition for any other other ones?
                            if (
                                len(partnerDefinition) == 1
                                and partner in prerequirements
                                and isActive(
                                    (partnerDefinition[0][0], partnerDefinition[0][1])
                                )
                            ):
                                # if so mark for deletion

                                indirect[molecule].add((partner, requirement))
                        else:
                            # is any single prerequirement a sufficient condition for any other other ones?
                            if (
                                len(partnerDefinition) == 1
                                and (
                                    partner,
                                    partnerDefinition[0][0],
                                    partnerDefinition[0][1],
                                )
                                in prerequirements
                            ):
                                # if so mark for deletion
                                indirect[molecule].add(
                                    (
                                        (
                                            partner,
                                            partnerDefinition[0][0],
                                            partnerDefinition[0][1],
                                        ),
                                        requirement,
                                    )
                                )

    for molecule in indirect:
        for indirectDependency in indirect[molecule]:
            dependencies[molecule][relationship].remove(indirectDependency)


def formatComponent(state, tense):
    if state[0] == 1:
        if tense == "past":
            compStr = "be bound"
        else:
            compStr = "bind"
    else:
        compStr = "be in state {0}".format(state[1])
    return compStr


def printDependencyLog(dependencies):
    log = StringIO()
    for molecule in dependencies:
        for requirementType in dependencies[molecule]:
            # if requirementType in ['independent']:
            #    continue
            for baseMolecule in dependencies[molecule][requirementType]:
                if requirementType == "requirement":
                    log.write(
                        "Molecule {0} needs component {1} to {2} for component {3} to {4}\n".format(
                            molecule,
                            baseMolecule[0][0],
                            formatComponent(baseMolecule[0][1:], "past"),
                            baseMolecule[1][0],
                            formatComponent(baseMolecule[1][1:], "present"),
                        )
                    )
                elif requirementType == "exclusion":
                    log.write(
                        "In molecule {0} component {1} is mutually exclusive of component {2}\n".format(
                            molecule, baseMolecule[0][0], baseMolecule[1][0]
                        )
                    )
                if requirementType == "nullrequirement":
                    pass
                # if requirementType == 'independent':
                #    log.write('The setting of {0} to {1} in molecule {3} is independent from {2}\n'.format(baseMolecule[0][0],baseMolecule[0][1],baseMolecule[1],molecule))
    return log.getvalue()


def removeCounter(requirementDependencies):
    finalDependencies = defaultdict(lambda: defaultdict(set))
    for requirement in requirementDependencies:
        req2 = requirement.split("%")[0]
        for dependencies in requirementDependencies[requirement]:
            finalDependencies[req2][dependencies] = finalDependencies[req2][
                dependencies
            ].union(requirementDependencies[requirement][dependencies])
    return finalDependencies


def getExclusionClusters(requirementDependencies):
    graphExclusionCliques = {x: [] for x in requirementDependencies.keys()}
    for molecule in requirementDependencies.keys():
        g = nx.Graph()
        gmod = nx.Graph()
        for exclusionRelationship in requirementDependencies[molecule]["exclusion"]:
            if not exclusionRelationship[0].lower().endswith(
                "mod"
            ) and not exclusionRelationship[1].lower().endswith("mod"):
                g.add_edge(exclusionRelationship[0], exclusionRelationship[1])
            elif exclusionRelationship[0].lower().endswith(
                "mod"
            ) and exclusionRelationship[1].lower().endswith("mod"):
                gmod.add_edge(exclusionRelationship[0], exclusionRelationship[1])
        for graph in [g, gmod]:
            if graph.nodes():
                while len(graph.nodes()) > 1:
                    cliques = sorted(
                        list(nx.find_cliques(graph)), key=len, reverse=True
                    )
                    if len(cliques[0]) > 1:
                        graphExclusionCliques[molecule].append(cliques[0])
                    else:
                        break
                    for node in cliques[0]:
                        graph.remove_node(node)
    return graphExclusionCliques


def getContextRequirements(
    inputfile, collapse=True, motifFlag=False, excludeReverse=False
):
    """
    Receives a BNG-XML file and returns the contextual dependencies implied by this file
    """
    molecules, rules, _ = readBNGXML.parseXML(inputfile)

    (
        label,
        center,
        context,
        product,
        atomicArray,
        actions,
        doubleAction,
    ) = extractCenterContext(rules, excludeReverse=excludeReverse)
    reactionCenterStateDictionary, doubleActionDict = getRestrictedChemicalStates(
        label, product, context, doubleAction
    )

    # molecule1 = 'socs3'
    # molecule2 = 'shp2'
    # print(molecule1,molecule2)
    # print('---',reactionCenterStateDictionary['gp130%0'][(molecule1,1,'')][molecule2])
    # print('---',reactionCenterStateDictionary['gp130%0'][(molecule2,1,'')][molecule1])
    # print(reactionCenterStateDictionary['STAT3%0'][('astmod',0,'AST')])
    # print('++++')

    # print(reactionCenterStateDictionary['Ras%0'][('Ras_GTPmod',0,'Ras_GTP')]['Ras_GDPmod'])
    backupstatedictionary = deepcopy(reactionCenterStateDictionary)
    # print(reactionCenterStateDictionary['EGFR%1'][('_Pmod',0,'_P')])
    # print(reactionCenterStateDictionary['EGFR%0'][('_Pmod',0,'_P')])
    # return
    # chemicalStates = getChemicalStates(rules)
    # totalStateDictionary = sortChemicalStates(chemicalStates)
    # print(reactionCenterStateDictionary['Shc%0'][('egfr', 0, '')]['mmod'])
    requirementDependencies = detectDependencies(
        reactionCenterStateDictionary, molecules
    )

    # print(requirementDependencies['JAK']['requirement'])
    # repression
    for molecule in reactionCenterStateDictionary:
        moleculeName = molecule.split("%")[0]
        for element in reactionCenterStateDictionary[molecule]:
            if not isActive(element[1:]):
                for partner in reactionCenterStateDictionary[molecule][element]:
                    for state in reactionCenterStateDictionary[molecule][element][
                        partner
                    ]:
                        if isActive(state):
                            repressiveState = (partner, state[0], state[1])
                            if (
                                reverseState(moleculeName, element, molecules),
                                repressiveState,
                            ) in requirementDependencies[moleculeName]["requirement"]:
                                # or (repressiveState, element[0]) in requirementDependencies[moleculeName]['independent']:
                                requirementDependencies[moleculeName]["repression"].add(
                                    (repressiveState[0], element[0])
                                )

    if motifFlag:
        requirementDependencies = getMotifRelationships(
            requirementDependencies, molecules
        )
        exclusionCliques = getExclusionClusters(requirementDependencies)
        removeIndirectDependencies(
            requirementDependencies, backupstatedictionary, motifFlag
        )

    else:
        removeIndirectDependencies(requirementDependencies, backupstatedictionary)
        getMutualExclusions(requirementDependencies, molecules)
        exclusionCliques = {}

    processNodes = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    if motifFlag:
        # double interactions
        print(doubleActionDict["gp130"])
        multiInteractionDict = defaultdict(
            lambda: defaultdict(lambda: defaultdict(list))
        )
        doubleInteractions = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        for molecule in [x for x in requirementDependencies if x in doubleActionDict]:
            for motif in requirementDependencies[molecule]:
                for relationship in requirementDependencies[molecule][motif]:

                    for combination1 in [True, False]:
                        for combination2 in [True, False]:
                            if (
                                len(
                                    doubleActionDict[molecule][relationship[0]][
                                        combination1
                                    ][relationship[1]][combination2]
                                )
                                > 0
                            ):
                                multiInteractionDict[molecule][
                                    (combination1, combination2)
                                ][relationship] = tuple(
                                    doubleActionDict[molecule][relationship[0]][
                                        combination1
                                    ][relationship[1]][combination2]
                                )
                                doubleInteractions[molecule][tuple(relationship)][
                                    motif
                                ].append((combination1, combination2))

        for molecule in doubleInteractions:
            for relationship in doubleInteractions[molecule]:
                for motif in doubleInteractions[molecule][relationship]:
                    if motif in [
                        "partialIndependence-",
                        "partialIndependence+",
                        "fullIndependence",
                    ]:
                        requirementDependencies[molecule][motif].remove(relationship)
                    for combination in doubleInteractions[molecule][relationship][
                        motif
                    ]:
                        label = multiInteractionDict[molecule][combination][
                            relationship
                        ]
                        if combination[0] and combination[1]:
                            requirementDependencies[molecule][
                                "doubleActivation"
                            ].append(relationship)
                            processNodes[molecule]["doubleActivation"][
                                relationship
                            ] = "{0}_{1}".format(molecule, "_".join(label))
                        elif not combination[0] and combination[1]:
                            if motif in ["ordering"]:
                                requirementDependencies[molecule][motif].remove(
                                    relationship
                                )
                            elif motif in ["repression"]:
                                if (
                                    relationship[1],
                                    relationship[0],
                                ) in requirementDependencies[molecule][motif]:
                                    requirementDependencies[molecule][motif].remove(
                                        (relationship[1], relationship[0])
                                    )
                            requirementDependencies[molecule]["reprordering"].append(
                                relationship
                            )
                            processNodes[molecule]["reprordering"][
                                relationship
                            ] = "{0}_{1}".format(molecule, "_".join(label))

                        elif not combination[0] and not combination[1]:
                            processNodes[molecule]["doubleRepression"][
                                relationship
                            ] = "{0}_{1}".format(molecule, "_".join(label))
                            if motif == "repression":
                                requirementDependencies[molecule][motif].remove(
                                    relationship
                                )
                                requirementDependencies[molecule][
                                    "doubleRepression"
                                ].append(relationship)
                            elif motif == "partialIndependence-":
                                # requirementDependencies[molecule][motif].remove(relationship)
                                requirementDependencies[molecule][
                                    "doubleRepression"
                                ].append(relationship)

    return (
        requirementDependencies,
        backupstatedictionary,
        exclusionCliques,
        processNodes,
    )


def reverseContextDict(dependencies):
    """
    returns a molecule/component pair/relationship hierarchy
    @param a molecule/relationship/component pair hierarchy
    """
    reverseDependencies = defaultdict(lambda: defaultdict(str))

    for molecule in dependencies:
        for dependencyType in dependencies[molecule]:
            if dependencyType == "independent":
                for relationship in dependencies[molecule][dependencyType]:
                    if relationship[0][1] == 1 or relationship[0][2] not in [
                        "0",
                        0,
                        "",
                    ]:
                        reverseDependencies[molecule][
                            (relationship[0][0], relationship[1])
                        ] = "independent"
            else:
                for relationship in dependencies[molecule][dependencyType]:

                    if dependencyType == "exclusion":
                        reverseDependencies[molecule][
                            (relationship[1][0], relationship[0][0])
                        ] = "nullrequirement"
                        reverseDependencies[molecule][
                            (relationship[0][0], relationship[1][0])
                        ] = "nullrequirement"
                    else:
                        reverseDependencies[molecule][
                            (relationship[1][0], relationship[0][0])
                        ] = dependencyType
    return reverseDependencies


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument("-i", "--input", type=str, help="settings file", required=True)
    return parser


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()
    inputFile = namespace.input
    # print(askQuestions(inputFile, 'EGFR', 'shc','grb2'))
    dependencies, backup, _, _ = getContextRequirements(
        inputFile, collapse=True, motifFlag=True
    )

    # print(dependencies)
    # print(dict(dependencies['EGFR']))
    # print(backup)
    # print(printDependencyLog(dependencies))



import colorlog

# global log level, can be set using
# from bionetgen.core.utils import logging
# logging.log_level = "DEBUG"
# options are "INFO","DEBUG","WARNING","ERROR","CRITICAL"
log_level = None

# set colorlog handler
fmter = colorlog.ColoredFormatter(
    "%(log_color)s%(levelname)s:%(name)s:%(message)s",
    log_colors={
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red",
    },
)


class BNGLogger:
    """
    Logger class that handles all logging in PyBNG. This class
    specifically tries to bridge the gap of using colorlog by
    itself and Cement frameworks own logging system.

    See [python logging](https://docs.python.org/3/howto/logging.html)
    for more information how logging in python works in general.

    Usage: BNGLogger()
           BNGLogger(app=app, level="DEBUG")

    Attributes
    ----------
    app : Cement App
        if using the logger with Cement framework, this will be the
        Cement app and each method will call app.log
    loc : str
        this attribute keeps track of the location string for better
        error reporting. for cement app this will be passed directly
        to app.log, for regular logging this will be parsed to get
        the right logger
    log_level : str
        this attribute sets the logging level to output. Options are
        the same as regular python logging, "DEBUG", "INFO", "WARNING",
        "ERROR", "CRITICAL"

    Methods
    -------
    get_logger : logger
        this method returns the correct logger after parsing the loc string
    debug : None
        same as all python logging, writes a message at DEBUG level
    info : None
        same as all python logging, writes a message at INFO level
    warning : None
        same as all python logging, writes a message at WARNING level
    error : None
        same as all python logging, writes a message at ERROR level
    critical : None
        same as all python logging, writes a message at CRITICAL level
    """

    def __init__(self, app=None, level="INFO", loc=None):
        self.app = app
        self.loc = loc
        # global ll overrides everything
        if log_level is not None:
            self.level = log_level
        # cli is second most important
        elif self.app is not None:
            if self.app.pargs.debug:
                self.level = "DEBUG"
                if self.level != self.app.log.get_level():
                    self.app.log.set_level(self.level)
            elif self.app.pargs.log_level is not None:
                self.level = app.pargs.log_level
                if self.level != self.app.log.get_level():
                    self.app.log.set_level(self.level)
        # what this is instantiated with is the least
        # at least for now
        else:
            self.level = level

    def get_logger(self, loc=None):
        """
        From a given loc string returns the correct python
        logger. Loc string should be of the form:

        'FILE_PATH : MODULE.METHOD'
        """
        if loc is None:
            if self.loc is None:
                logger = colorlog.getLogger()
            else:
                loc_full = self.loc.split(":")[-1]
                module_name = loc_full.split(".")[0].strip()
                logger = colorlog.getLogger(module_name)
        else:
            loc_full = loc.split(":")[-1]
            module_name = loc_full.split(".")[0].strip()
            logger = colorlog.getLogger(module_name)
        if not logger.hasHandlers():
            handler = colorlog.StreamHandler()
            handler.setFormatter(fmter)
            logger.addHandler(handler)
        logger.setLevel(self.level)
        return logger

    def debug(self, msg, loc=None):
        """
        Debug level messages
        """
        if self.app is None:
            logger = self.get_logger(loc=loc)
            if loc is not None:
                msg = f"from: {loc} : {msg}"
            elif self.loc is not None:
                msg = f"from: {self.loc} : {msg}"
            logger.debug(msg)
        else:
            self.app.log.debug(msg, loc)

    def info(self, msg, loc=None):
        """
        Info level messages
        """
        if self.app is None:
            logger = self.get_logger(loc=loc)
            if loc is not None:
                msg = f"from: {loc} : {msg}"
            elif self.loc is not None:
                msg = f"from: {self.loc} : {msg}"
            logger.info(msg)
        else:
            self.app.log.info(msg, loc)

    def warning(self, msg, loc=None):
        """
        Warning level messages
        """
        if self.app is None:
            logger = self.get_logger(loc=loc)
            if loc is not None:
                msg = f"from: {loc} : {msg}"
            elif self.loc is not None:
                msg = f"from: {self.loc} : {msg}"
            logger.warning(msg)
        else:
            self.app.log.warning(msg, loc)

    def error(self, msg, loc=None):
        """
        Error level messages
        """
        if self.app is None:
            logger = self.get_logger(loc=loc)
            if loc is not None:
                msg = f"from: {loc} : {msg}"
            elif self.loc is not None:
                msg = f"from: {self.loc} : {msg}"
            logger.error(msg)
        else:
            self.app.log.error(msg, loc)

    def critical(self, msg, loc=None):
        """
        Critical level messages
        """
        if self.app is None:
            logger = self.get_logger(loc=loc)
            if loc is not None:
                msg = f"from: {loc} : {msg}"
            elif self.loc is not None:
                msg = f"from: {self.loc} : {msg}"
            logger.critical(msg)
        else:
            self.app.log.critical(msg, loc)



import xmltodict, re

from bionetgen.main import BioNetGen
from bionetgen.core.exc import BNGParseError, BNGModelError
from tempfile import TemporaryFile

from .bngfile import BNGFile
from .xmlparsers import ParameterBlockXML, CompartmentBlockXML, ObservableBlockXML
from .xmlparsers import SpeciesBlockXML, MoleculeTypeBlockXML, FunctionBlockXML
from .xmlparsers import RuleBlockXML, EnergyPatternBlockXML, PopulationMapBlockXML
from .blocks import ActionBlock
from bionetgen.core.utils.utils import ActionList

# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]


class BNGParser:
    """
    Parser object that deals with reading in the BNGL file and
    setting up the model object

    Usage: BNGParser(bngl_path)
           BNGParser(bngl_path, BNGPATH)

    Attributes
    ----------
    bngfile : BNGFile
        BNGFile object that's responsible for .bngl file manipulations

    Methods
    -------
    parse_model(model_file)
        parses the BNGL model at the given path and adds everything to a given model object
    parse_xml(xml_str)
        parses given xml string and adds everything to a given model object
    """

    def __init__(
        self, path, BNGPATH=def_bng_path, parse_actions=True, generate_network=False
    ) -> None:
        self.to_parse_actions = parse_actions
        self.bngfile = BNGFile(path, generate_network=generate_network)
        self.alist = ActionList()
        self.alist.define_parser()

    def parse_model(self, model_obj) -> None:
        """
        Will determine the parser route eventually and call the right
        parser
        """
        self._parse_model_bngpl(model_obj)
        if self.to_parse_actions:
            self.parse_actions(model_obj)

    def _parse_model_bngpl(self, model_obj) -> None:
        # get file path
        model_file = self.bngfile.path

        # this route runs BNG2.pl on the bngl and parses
        # the XML instead
        if model_file.endswith(".bngl"):
            # TODO: Add verbosity option to the library
            # print("Attempting to generate XML")
            with TemporaryFile("w+") as xml_file:
                if self.bngfile.generate_xml(xml_file):
                    # TODO: Add verbosity option to the library
                    xmlstr = xml_file.read()
                    # < is not a valid XML character, we need to replace it
                    xmlstr = xmlstr.replace('relation="<', 'relation="&lt;')
                    self.parse_xml(xmlstr, model_obj)
                    model_obj.reset_compilation_tags()
                else:
                    raise BNGModelError(
                        self.bngfile.path, message="XML file couldn't be generated"
                    )
        elif model_file.endswith(".xml"):
            with open(model_file, "r") as f:
                xml_str = f.read()
                # < is not a valid XML character, we need to replace it
                xmlstr = xml_str.replace('relation="<', 'relation="&lt;')
                self.parse_xml(xml_str, model_obj)
            model_obj.reset_compilation_tags()
        else:
            raise NotImplementedError(
                "The extension of {} is not supported".format(model_file)
            )

    def parse_actions(self, model_obj):
        if len(self.bngfile.parsed_actions) > 0:
            ablock = ActionBlock()
            # we have actions in file, let's get them
            # import ipdb;ipdb.set_trace()
            left = []
            for action in self.bngfile.parsed_actions:
                # some cleanup, first we remove comments
                action = re.sub(r"\#.*", "", action)
                # now we remove whitespaces
                action = re.sub(r"\s", "", action)
                # if we don't have anything left, move on
                if len(action) == 0:
                    continue
                # use pyparsing for parsing the action into a list
                try:
                    action_list = self.alist.action_parser.parseString(action)
                except Exception as e:
                    raise BNGParseError(
                        self.bngfile.path, f"Failed to parse action {action}"
                    )
                # we could have ";" in the action, so we need to remove it
                if action_list[-1] == ";":
                    _ = action_list.pop(-1)
                # we we move onto actually making the action object
                # first value is always the action type, remove
                atype = action_list.pop(0)
                # all actions have "()", remove
                action_list = action_list[1:-1]
                # be done if we don't have anything left
                if len(action_list) == 0:
                    # we don't have any arguments
                    ablock.add_action(atype, {})
                    continue
                # we have arguments now onto argument parsing
                # we check the action type and process accordingly
                if atype in self.alist.no_setter_syntax:
                    # these are actions like setParameter("test", 10), setModelName("name")
                    if len(action_list) == 1:
                        # this is of the form action("argument")
                        ablock.add_action(atype, {action_list[0]: None})
                        continue
                    elif len(action_list) == 3:
                        # TODO: Error checking here!
                        if action_list[1] == ",":
                            # this is of the form action(argument, value)
                            ablock.add_action(
                                atype, {action_list[0]: None, action_list[2]: None}
                            )
                            continue
                elif atype in self.alist.square_braces:
                    # these are actions like saveParameters(["a","b"])
                    # TODO: Error checking here!
                    if action_list[0] == "[":
                        # remove square braces
                        action_list = action_list[1:-1]
                    arg_dict = {}
                    for arg in action_list:
                        arg_dict[arg] = None
                    ablock.add_action(atype, arg_dict)
                    continue
                elif atype in self.alist.normal_types:
                    # finally a normal action, we have {} and => syntax
                    # TODO: Error checking here!
                    if action_list[0] == "{":
                        # remove curly braces
                        action_list = action_list[1:-1]
                    arg_dict = {}
                    if len(action_list) == 0:
                        ablock.add_action(atype, arg_dict)
                        continue
                    while len(action_list) > 0:
                        arg_name = action_list.pop(0)
                        connector = action_list.pop(0)
                        if connector != "=>":
                            raise BNGParseError(
                                self.bngfile.path, f"Action {action} is malformed"
                            )
                        if arg_name in self.alist.irregular_args:
                            arg_type = self.alist.irregular_args[arg_name]
                            if arg_type == "dict":
                                # process dict
                                start_curly = action_list.pop(0)
                                # make sure we are actually reading a dict
                                if start_curly != "{":
                                    raise BNGParseError(
                                        self.bngfile.path,
                                        f"Action {action} is malformed",
                                    )
                                value_str = "{"
                                end_curly = None
                                while end_curly is None:
                                    # we are looping over A, =>, B and want to
                                    # generate { A=>B, C=>D, etc }
                                    dict_key = action_list.pop(0)
                                    if dict_key == "}":
                                        # we are done
                                        end_curly = dict_key
                                    else:
                                        if len(value_str) > 1:
                                            value_str += ","
                                        dict_conn = action_list.pop(0)
                                        dict_val = action_list.pop(0)
                                        if dict_conn != "=>":
                                            raise BNGParseError(
                                                self.bngfile.path,
                                                f"Action {action} is malformed",
                                            )
                                        value_str += dict_key + dict_conn + dict_val
                                value_str += "}"
                                arg_value = value_str
                            elif arg_type == "list":
                                # process list
                                start_curly = action_list.pop(0)
                                # make sure we are actually reading a dict
                                if start_curly != "[":
                                    raise BNGParseError(
                                        self.bngfile.path,
                                        f"Action {action} is malformed",
                                    )
                                value_str = "["
                                end_curly = None
                                while end_curly is None:
                                    argument_element = action_list.pop(0)
                                    if argument_element == "]":
                                        end_curly = argument_element
                                    else:
                                        if len(value_str) > 1:
                                            value_str += ","
                                        value_str += argument_element
                                value_str += "]"
                                arg_value = value_str
                        else:
                            arg_value = action_list.pop(0)
                        arg_dict[arg_name] = arg_value
                    ablock.add_action(atype, arg_dict)
                    continue
                else:
                    raise BNGParseError(
                        self.bngfile.path, f"Action type {atype} is not recognized."
                    )
            model_obj.add_block(ablock)

    def parse_xml(self, xml_str, model_obj) -> None:
        xml_dict = xmltodict.parse(xml_str)
        # catch non-BNG XML files
        if "sbml" not in xml_dict:
            if "model" not in xml_dict["sbml"]:
                raise BNGParseError(
                    self.bngfile.path,
                    "Input model is invalid. Please ensure model is in proper BNGL or BNG-XML format",
                )
        model_obj.xml_dict = xml_dict
        first_key = list(xml_dict.keys())[0]
        xml_model = xml_dict[first_key]["model"]
        model_obj.model_name = xml_model["@id"]
        for listkey in xml_model.keys():
            if listkey == "ListOfParameters":
                param_list = xml_model[listkey]
                if param_list is not None:
                    params = param_list["Parameter"]
                    xml_parser = ParameterBlockXML(params)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfObservables":
                obs_list = xml_model[listkey]
                if obs_list is not None:
                    obs = obs_list["Observable"]
                    xml_parser = ObservableBlockXML(obs)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfCompartments":
                comp_list = xml_model[listkey]
                if comp_list is not None:
                    comps = comp_list["compartment"]
                    xml_parser = CompartmentBlockXML(comps)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfMoleculeTypes":
                mtypes_list = xml_model[listkey]
                if mtypes_list is not None:
                    mtypes = mtypes_list["MoleculeType"]
                    xml_parser = MoleculeTypeBlockXML(mtypes)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfSpecies":
                species_list = xml_model[listkey]
                if species_list is not None:
                    species = species_list["Species"]
                    xml_parser = SpeciesBlockXML(species)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfReactionRules":
                rrules_list = xml_model[listkey]
                if rrules_list is not None:
                    rrules = rrules_list["ReactionRule"]
                    xml_parser = RuleBlockXML(rrules)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfFunctions":
                # TODO: Optional expression parsing?
                # TODO: Add arguments correctly
                func_list = xml_model[listkey]
                if func_list is not None:
                    funcs = func_list["Function"]
                    xml_parser = FunctionBlockXML(funcs)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfEnergyPatterns":
                ep_list = xml_model[listkey]
                if ep_list is not None:
                    eps = ep_list["EnergyPattern"]
                    xml_parser = EnergyPatternBlockXML(eps)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfPopulationMaps":
                pm_list = xml_model[listkey]
                if pm_list is not None:
                    pms = pm_list["PopulationMap"]
                    xml_parser = PopulationMapBlockXML(pms)
                    model_obj.add_block(xml_parser.parsed_obj)
        # And that's the end of parsing
        # TODO: Add verbosity option to the library
        # print("Parsing complete")



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 19 14:28:16 2012

@author: proto
"""
from lxml import etree
from . import smallStructures as st
from io import StringIO

# http://igraph.sourceforge.net/documentation.html
# ----------------------------------------------------------------------


def findBond(bondDefinitions, component):
    """
    Returns an appropiate bond number when veryfying how
    to molecules connect in a species
    """
    for idx, bond in enumerate(bondDefinitions.getchildren()):
        if component in [bond.get("site1"), bond.get("site2")]:
            return str(idx)


def createMolecule(molecule, bonds):
    nameDict = {}
    mol = st.Molecule(molecule.get("name"), molecule.get("id"))
    if molecule.get("compartment") not in ["", None]:
        mol.setCompartment(molecule.get("compartment"))
    nameDict[molecule.get("id")] = molecule.get("name")
    listOfComponents = molecule.find(
        ".//{http://www.sbml.org/sbml/level3}ListOfComponents"
    )
    if listOfComponents is not None:
        for element in listOfComponents:
            component = st.Component(element.get("name"), element.get("id"))
            nameDict[element.get("id")] = element.get("name")
            if element.get("numberOfBonds") in ["+", "?"]:
                component.addBond(element.get("numberOfBonds"))
            elif element.get("numberOfBonds") != "0":
                component.addBond(findBond(bonds, element.get("id")))
            state = element.get("state") if element.get("state") != None else ""
            component.states.append(state)
            component.activeState = state
            mol.addComponent(component)
    return mol, nameDict


def createSpecies(pattern):
    tmpDict = {}
    species = st.Species()
    species.idx = pattern.get("id")
    mol = pattern.find(".//{http://www.sbml.org/sbml/level3}ListOfMolecules")
    bonds = pattern.find(".//{http://www.sbml.org/sbml/level3}ListOfBonds")
    for molecule in mol.getchildren():
        molecule, nameDict = createMolecule(molecule, bonds)
        tmpDict.update(nameDict)
        species.addMolecule(molecule)
        if bonds != None:
            species.bonds = [(bond.get("site1"), bond.get("site2")) for bond in bonds]
        tmpDict.update(nameDict)
    return species, tmpDict


def parseRule(rule, parameterDict):
    """
    Parses a rule XML section
    Returns: a list of the reactants and products used, followed by the mapping
    between the two and the list of operations that were performed
    """
    rp = rule.find(".//{http://www.sbml.org/sbml/level3}ListOfReactantPatterns")
    pp = rule.find(".//{http://www.sbml.org/sbml/level3}ListOfProductPatterns")
    mp = rule.find(".//{http://www.sbml.org/sbml/level3}Map")
    op = rule.find(".//{http://www.sbml.org/sbml/level3}ListOfOperations")
    rt = rule.find(".//{http://www.sbml.org/sbml/level3}RateLaw")
    nameDict = {}
    reactants = []
    products = []
    actions = []
    mappings = []

    if len(rp) == 0:
        sp = st.Species()
        ml = st.Molecule("0", "")
        sp.addMolecule(ml)
        reactants.append(sp)
    if len(pp) == 0:
        sp = st.Species()
        ml = st.Molecule("0", "")
        sp.addMolecule(ml)
        products.append(sp)
    for pattern in rp:
        elm, tmpDict = createSpecies(pattern)
        reactants.append(elm)
        nameDict.update(tmpDict)
    for pattern in pp:
        elm, tmpDict = createSpecies(pattern)
        products.append(elm)
        nameDict.update(tmpDict)
    for operation in op:
        action = st.Action()
        tag = operation.tag
        tag = tag.replace("{http://www.sbml.org/sbml/level3}", "")
        if operation.get("site1") != None:
            action.setAction(tag, operation.get("site1"), operation.get("site2"))
        else:
            action.setAction(tag, operation.get("site"), None)
        actions.append(action)
    for mapping in mp:
        tmpMap = (mapping.get("sourceID"), mapping.get("targetID"))
        mappings.append(tmpMap)
    rateConstants = rt.find(".//{http://www.sbml.org/sbml/level3}ListOfRateConstants")
    if rateConstants == None:
        rateConstants = rt.get("name")
    else:
        for constant in rateConstants:
            tmp = constant.get("value")
        rateConstants = tmp
    rateConstantsValue = (
        parameterDict[rateConstants]
        if rateConstants in parameterDict
        else rateConstants
    )
    # rule = st.Rule()
    label = rule.get("name")
    label = label.replace("(", "_").replace(")", "_")
    rule = st.Rule(label)
    rule.addReactantList(reactants)
    rule.addProductList(products)
    rule.addActionList(actions)
    rule.addMappingList(mappings)
    rule.addRate(rateConstants)

    # return reactants, products, actions, mappings, nameDict,rateConstantsValue,rateConstants
    return rule, nameDict, rateConstantsValue, rateConstants


def parseMolecules(molecules):
    """
    Parses an XML molecule section
    Returns: a molecule structure
    """

    mol = st.Molecule(molecules.get("id"), molecules.get("id"))
    components = molecules.find(
        ".//{http://www.sbml.org/sbml/level3}ListOfComponentTypes"
    )
    if components != None:
        for component in components.getchildren():
            comp = parseComponent(component)
            mol.addComponent(comp)
    return mol


def parseComponent(component):
    """
    parses  a bngxml molecule types section
    """
    comp = st.Component(component.get("id"), component.get("id"))
    states = component.find(".//{http://www.sbml.org/sbml/level3}ListOfAllowedStates")
    if states != None:
        for state in states.getchildren():
            comp.addState(state.get("id"))
    return comp


def parseObservable(observable):
    nameDict = {}
    name = observable.get("name")
    otype = observable.get("type")
    rp = observable.find(".//{http://www.sbml.org/sbml/level3}ListOfPatterns")

    patternList = []
    for pattern in rp:
        elm, tmpDict = createSpecies(pattern)
        patternList.append(elm)

    return name, otype, patternList


def parseObservables(observables):
    observableDescription = []

    for observable in observables:
        observableDescription.append(parseObservable(observable))
    return observableDescription


def parseFunction(function):
    referenceList = []
    name = function.get("id")
    # expression = function.find('.//{http://www.sbml.org/sbml/level3}Expression')
    expression = function.findtext(".//{http://www.sbml.org/sbml/level3}Expression")

    references = function.find(".//{http://www.sbml.org/sbml/level3}ListOfReferences")
    for reference in references:
        referenceList.append([reference.get("name"), reference.get("type")])

    return name, expression, referenceList


def parseFunctions(functions):

    functionDescription = []

    for function in functions:
        functionDescription.append(parseFunction(function))
    return functionDescription


def parseFullXML(xmlFile):
    doc = etree.parse(xmlFile)
    molecules = doc.findall(".//{http://www.sbml.org/sbml/level3}MoleculeType")
    seedspecies = doc.findall(".//{http://www.sbml.org/sbml/level3}Species")
    rules = doc.findall(".//{http://www.sbml.org/sbml/level3}ReactionRule")
    functions = doc.findall(".//{http://www.sbml.org/sbml/level3}Function")

    structureDefinitions = {}
    ruleDescription = []
    moleculeList = []
    seedSpeciesList = []
    compartmentList = []
    observables = doc.findall(".//{http://www.sbml.org/sbml/level3}Observable")
    parameters = doc.findall(".//{http://www.sbml.org/sbml/level3}Parameter")
    compartments = doc.findall(".//{http://www.sbml.org/sbml/level3}compartment")

    parameterDict = {}
    for parameter in parameters:
        parameterDict[parameter.get("id")] = parameter.get("value")

    structureDefinitions["parameters"] = parameterDict

    for compartment in compartments:
        compartmentDict = {}
        compartmentDict["identifier"] = compartment.get("id")
        compartmentDict["dimensions"] = compartment.get("spatialDimensions")
        compartmentDict["size"] = compartment.get("size")
        compartmentDict["outside"] = compartment.get("outside")
        compartmentList.append(compartmentDict)

    structureDefinitions["compartments"] = compartmentList

    for molecule in molecules:
        moleculeList.append(parseMolecules(molecule))

    structureDefinitions["molecules"] = moleculeList

    for species in seedspecies:
        concentration = species.get("concentration")
        compartment = species.get("compartment")
        seedSpeciesList.append(
            {
                "concentration": concentration,
                "structure": createSpecies(species)[0],
                "compartment": compartment,
            }
        )

    structureDefinitions["seedspecies"] = seedSpeciesList

    for rule in rules:
        # description = parseRule(rule, parameterDict)
        # if 'reverse' in description[0].label:
        #    ruleDescription[-1][0].bidirectional= True
        #    ruleDescription[-1][0].rates.append(description[0].rates[0])
        # else:
        ruleDescription.append(parseRule(rule, parameterDict))

    structureDefinitions["rules"] = ruleDescription
    structureDefinitions["functions"] = parseFunctions(functions)
    structureDefinitions["observables"] = parseObservables(observables)

    return structureDefinitions


def parseXMLStruct(doc):
    molecules = doc.findall(".//{http://www.sbml.org/sbml/level3}MoleculeType")
    rules = doc.findall(".//{http://www.sbml.org/sbml/level3}ReactionRule")
    ruleDescription = []
    moleculeList = []

    parameters = doc.findall(".//{http://www.sbml.org/sbml/level3}Parameter")
    parameterDict = {}
    for parameter in parameters:
        parameterDict[parameter.get("id")] = parameter.get("value")

    for molecule in molecules:
        moleculeList.append(parseMolecules(molecule))

    for rule in rules:
        description = parseRule(rule, parameterDict)
        # if 'reverse' in description[0].label:
        #    ruleDescription[-1][0].bidirectional= True
        #    ruleDescription[-1][0].rates.append(description[0].rates[0])
        # else:
        ruleDescription.append(parseRule(rule, parameterDict))
    return moleculeList, ruleDescription, parameterDict


def parseXMLFromString(xmlString):
    doc = etree.fromstring(xmlString)
    return parseXMLStruct(doc)


def parseFullXMLFromString(xmlString):
    doc = etree.fromstring(xmlString)
    return parseFullXML(doc)


def parseXML(xmlFile):
    doc = etree.parse(xmlFile)
    return parseXMLStruct(doc)


def getNumObservablesXML(xmlFile):
    doc = etree.parse(xmlFile)
    observables = doc.findall(".//{http://www.sbml.org/sbml/level3}Observable")
    return len(observables)


def createBNGLFromDescription(namespace):
    """creates a bngl file from a dictionary description containing molecules, species, etc."""

    bnglString = StringIO()

    bnglString.write("begin model\n")
    # parameters
    bnglString.write("begin parameters\n")
    for parameterName in namespace["parameters"]:
        bnglString.write(
            "\t{0} {1}\n".format(parameterName, namespace["parameters"][parameterName])
        )
    bnglString.write("end parameters\n")

    # molecule types
    bnglString.write("begin molecule types\n")
    for molecule in namespace["molecules"]:
        bnglString.write("\t" + molecule.str2() + "\n")
    bnglString.write("end molecule types\n")

    # compartments
    bnglString.write("begin compartments\n")
    for compartment in namespace["compartments"]:
        bnglString.write(
            "\t{0} {1} {2}".format(
                compartment["identifier"],
                compartment["dimensions"],
                compartment["size"],
            )
        )
        if compartment["outside"]:
            bnglString.write("\t {0}\n".format(compartment["outside"]))
        else:
            bnglString.write("\n")
    bnglString.write("end compartments\n")

    # seed species
    bnglString.write("begin seed species\n")
    for seedspecies in namespace["seedspecies"]:
        if seedspecies["concentration"] not in [0, "0", "0.0"]:
            bnglString.write(
                "\t{0} {1}\n".format(
                    seedspecies["structure"], seedspecies["concentration"]
                )
            )
    bnglString.write("end seed species\n")

    # observables
    bnglString.write("begin observables\n")
    for observables in namespace["observables"]:
        bnglString.write(
            "\t{0} {1} {2}\n".format(
                observables[1],
                observables[0],
                ", ".join([str(x) for x in observables[2]]),
            )
        )
    bnglString.write("end observables\n")

    # functions
    bnglString.write("begin functions\n")
    for functions in namespace["functions"]:
        bnglString.write("\t{0}() = {1}\n".format(functions[0], functions[1]))
    bnglString.write("end functions\n")

    # rules
    bnglString.write("begin reaction rules\n")
    for rule in namespace["rules"]:
        bnglString.write("\t{0}\n".format(rule[0]))
    bnglString.write("end reaction rules\n")

    bnglString.write("end model\n")

    return bnglString.getvalue()


if __name__ == "__main__":
    # mol,rule,par = parseXML("output19.xml")
    # print([str(x) for x in mol])
    with open("output19.xml", "r") as f:
        s = f.read()
    print(parseFullXML("output19.xml")["rules"][0][0].actions[0].action)
    # print(getNumObservablesXML('output19.xml'))
